from __future__ import annotations

import abc
import contextlib
import decimal
from abc import ABC
from typing import ClassVar, Protocol

import attrs

from caqtus.device.runtime import Device
from caqtus.shot_compilation.timed_instructions import TimedInstruction
from ..timing import TimeStep
from ..trigger import Trigger, is_trigger


@attrs.define(slots=False)
class Sequencer(Device, ABC):
    """Abstract base class for a sequencer device.

    This function defines the methods that a sequencer device must implement to be
    compatible with the caqtus framework.

    Attributes:
        time_step: The time step of the sequencer in nanoseconds.
            This value cannot be changed after the sequencer has been created.
        trigger: Indicates how the sequence is started and how it is clocked.
            This value cannot be changed after the sequencer has been created.
    """

    channel_number: ClassVar[int]

    time_step: TimeStep = attrs.field(on_setattr=attrs.setters.frozen)
    trigger: Trigger = attrs.field(on_setattr=attrs.setters.frozen)

    @time_step.validator  # type: ignore
    def _validate_time_step(self, _, value):
        if not isinstance(value, decimal.Decimal):
            raise TypeError("Time step must be a decimal number")
        if value <= 0:
            raise ValueError("Time step must be greater than zero")

    @trigger.validator  # type: ignore
    def _validate_trigger(self, _, value):
        if not is_trigger(value):
            raise ValueError(f"Invalid trigger {value}")

    @abc.abstractmethod
    def program_sequence(self, sequence: TimedInstruction) -> ProgrammedSequence:
        """Program the sequence into the device.

        This method just writes the sequence to the device. It does not start the
        sequence.

        Args:
            sequence: The sequence to be programmed into the sequencer.

        Returns:
            An object that can be used to start the sequence.
        """

        raise NotImplementedError


class SequenceStatus(Protocol):
    """A protocol that defines the interface to check the status of a sequence."""

    @abc.abstractmethod
    def is_finished(self) -> bool:
        raise NotImplementedError


class ProgrammedSequence(Protocol):
    """A protocol that defines the interface to start a sequence."""

    @abc.abstractmethod
    def run(self) -> contextlib.AbstractContextManager[SequenceStatus]:
        """Start the sequence.

        Returns:
            A context manager that can be used to start a sequence.

            The sequence must be started when the context manager is entered.
            The context manager must yield a status object that can be used to check
            the progress of the sequence.

            When the context manager is exited, the sequence should be stopped.
            The context manager must not be exited without error if the sequence is
            still running.
        """

        raise NotImplementedError



"""Automatically generates GUI components for attrs classes."""

from ._device_config_editor import (
    build_device_configuration_editor,
    get_editor_builder,
    generate_device_configuration_editor,
)
from ._editor_builder import (
    EditorBuilder,
    TypeNotRegisteredError,
    build_attrs_class_editor,
    build_literal_editor,
    generate_enum_editor,
    AttributeOverride,
)
from ._int_editor import IntegerEditor
from ._string_editor import StringEditor
from ._value_editor import ValueEditor


__all__ = [
    "generate_device_configuration_editor",
    "build_device_configuration_editor",
    "build_attrs_class_editor",
    "build_literal_editor",
    "generate_enum_editor",
    "get_editor_builder",
    "EditorBuilder",
    "IntegerEditor",
    "StringEditor",
    "ValueEditor",
    "TypeNotRegisteredError",
    "AttributeOverride",
]



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'graphplot_main_window.ui'
##
## Created by: Qt User Interface Compiler version 6.6.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (
    QCoreApplication,
    QDate,
    QDateTime,
    QLocale,
    QMetaObject,
    QObject,
    QPoint,
    QRect,
    QSize,
    QTime,
    QUrl,
    Qt,
)
from PySide6.QtGui import (
    QAction,
    QBrush,
    QColor,
    QConicalGradient,
    QCursor,
    QFont,
    QFontDatabase,
    QGradient,
    QIcon,
    QImage,
    QKeySequence,
    QLinearGradient,
    QPainter,
    QPalette,
    QPixmap,
    QRadialGradient,
    QTransform,
)
from PySide6.QtWidgets import (
    QApplication,
    QMainWindow,
    QMenu,
    QMenuBar,
    QSizePolicy,
    QStatusBar,
    QWidget,
)


class Ui_GraphPlotMainWindow(object):
    def setupUi(self, GraphPlotMainWindow):
        if not GraphPlotMainWindow.objectName():
            GraphPlotMainWindow.setObjectName("GraphPlotMainWindow")
        GraphPlotMainWindow.resize(800, 600)
        self.centralwidget = QWidget(GraphPlotMainWindow)
        self.centralwidget.setObjectName("centralwidget")
        GraphPlotMainWindow.setCentralWidget(self.centralwidget)
        self.menubar = QMenuBar(GraphPlotMainWindow)
        self.menubar.setObjectName("menubar")
        self.menubar.setGeometry(QRect(0, 0, 800, 21))
        self.docks_menu = QMenu(self.menubar)
        self.docks_menu.setObjectName("docks_menu")
        GraphPlotMainWindow.setMenuBar(self.menubar)
        self.statusbar = QStatusBar(GraphPlotMainWindow)
        self.statusbar.setObjectName("statusbar")
        GraphPlotMainWindow.setStatusBar(self.statusbar)

        self.menubar.addAction(self.docks_menu.menuAction())

        self.retranslateUi(GraphPlotMainWindow)

        QMetaObject.connectSlotsByName(GraphPlotMainWindow)

    # setupUi

    def retranslateUi(self, GraphPlotMainWindow):
        GraphPlotMainWindow.setWindowTitle(
            QCoreApplication.translate("GraphPlotMainWindow", "MainWindow", None)
        )
        self.docks_menu.setTitle(
            QCoreApplication.translate("GraphPlotMainWindow", "Docks", None)
        )

    # retranslateUi



from caqtus.gui.condetrol.timelanes_editor import TimeLaneModel
from caqtus.gui.condetrol.timelanes_editor.camera_lane_editor import CameraTimeLaneModel
from caqtus.types.timelane import CameraTimeLane
from caqtus.utils import serialization
from caqtus.utils.serialization import JSON
from ._extension import TimeLaneExtension


def create_camera_lane(number_steps: int) -> CameraTimeLane:
    return CameraTimeLane([None] * number_steps)


def create_lane_model(lane, lane_name: str) -> TimeLaneModel:
    model = CameraTimeLaneModel(lane_name)
    model.set_lane(lane)
    return model


def create_lane_delegate(lane, lane_name: str) -> None:
    return None


def dump_camera_lane(time_lane: CameraTimeLane):
    return serialization.converters["json"].unstructure(time_lane, CameraTimeLane)


def load_camera_lane(content: JSON):
    return serialization.converters["json"].structure(content, CameraTimeLane)


camera_time_lane_extension = TimeLaneExtension(
    label="Camera",
    lane_type=CameraTimeLane,
    dumper=dump_camera_lane,
    loader=load_camera_lane,
    lane_factory=create_camera_lane,
    lane_model_factory=create_lane_model,
    lane_delegate_factory=create_lane_delegate,
    type_tag="camera",
)



from __future__ import annotations

import abc
import functools
import math
from collections.abc import Iterable, Sequence
from typing import Optional, Mapping, Any

import attrs
import numpy as np

from caqtus.shot_compilation import ShotContext
from caqtus.shot_compilation.lane_compilation import DimensionedSeries
from caqtus.shot_compilation.timed_instructions import (
    TimedInstruction,
    Pattern,
    Concatenated,
    concatenate,
    Repeated,
    Ramp,
    create_ramp,
)
from caqtus.types.units import Unit, InvalidDimensionalityError, Quantity, dimensionless
from caqtus.types.variable_name import DottedVariableName
from caqtus.utils.itertools import pairwise
from .channel_output import ChannelOutput
from ..timing import TimeStep


class TimeIndependentMapping(ChannelOutput, abc.ABC):
    """A functional mapping of input values to output values independent of time.

    This represents channel transformations of the form:

    .. math::
        y(t) = f(x_0(t), x_1(t), ..., x_n(t))

    where x_0, x_1, ..., x_n are the input and y is the output.
    """

    @abc.abstractmethod
    def inputs(self) -> tuple[ChannelOutput, ...]:
        """Returns the input values of the mapping."""

        raise NotImplementedError

    def evaluate_max_advance_and_delay(
        self,
        time_step: TimeStep,
        variables: Mapping[DottedVariableName, Any],
    ) -> tuple[int, int]:
        advances_and_delays = [
            input_.evaluate_max_advance_and_delay(time_step, variables)
            for input_ in self.inputs()
        ]
        advances, delays = zip(*advances_and_delays, strict=True)
        return max(advances), max(delays)


def data_points_converter(data_points: Iterable[tuple[float, float]]):
    point_to_tuple = [(x, y) for x, y in data_points]
    return tuple(sorted(point_to_tuple))


@attrs.define
class CalibratedAnalogMapping(TimeIndependentMapping):
    """Maps its input to an output quantity by interpolating a set of points.

    This mapping is useful for example when one needs to convert an experimentally
    measurable quantity (e.g. the frequency sent to an AOM) as a function of a control
    parameter (e.g. the voltage sent to the AOM driver).
    In this example, we need to know which voltage to apply to the AOM driver to obtain
    a given frequency.
    This conversion is defined by a set of points (x, y) where x is the input quantity
    and y is the output quantity.
    In the example above, x would be the frequency and y would be the voltage, because
    for a given frequency, we need to know which voltage to apply to the AOM driver.

    Attributes:
        input_units: The units of the input quantity
        input_: Describe the input argument of the mapping.
        output_units: The units of the output quantity
        measured_data_points: tuple of (input, output) tuples.
        The points will be rearranged to have the inputs sorted.
    """

    input_: ChannelOutput = attrs.field(
        validator=attrs.validators.instance_of(ChannelOutput),
        on_setattr=attrs.setters.validate,
    )
    input_units: Optional[str] = attrs.field(
        converter=attrs.converters.optional(str),
        on_setattr=attrs.setters.convert,
    )
    output_units: Optional[str] = attrs.field(
        converter=attrs.converters.optional(str),
        on_setattr=attrs.setters.convert,
    )
    measured_data_points: tuple[tuple[float, float], ...] = attrs.field(
        converter=data_points_converter, on_setattr=attrs.setters.convert
    )

    @property
    def input_values(self) -> tuple[float, ...]:
        return tuple(x[0] for x in self.measured_data_points)

    @property
    def output_values(self) -> tuple[float, ...]:
        return tuple(x[1] for x in self.measured_data_points)

    def inputs(self) -> tuple[ChannelOutput]:
        return (self.input_,)

    def __getitem__(self, index: int) -> tuple[float, float]:
        return self.measured_data_points[index]

    def __setitem__(self, index: int, values: tuple[float, float]):
        new_data_points = list(self.measured_data_points)
        new_data_points[index] = values
        self.measured_data_points = tuple(new_data_points)

    def set_input(self, index: int, value: float):
        self[index] = (value, self[index][1])

    def set_output(self, index: int, value: float):
        self[index] = (self[index][0], value)

    def pop(self, index: int):
        """Remove a data point from the mapping."""

        new_data_points = list(self.measured_data_points)
        new_data_points.pop(index)
        self.measured_data_points = tuple(new_data_points)

    def insert(self, index: int, input_: float, output: float):
        """Insert a data point into the mapping."""

        new_data_points = list(self.measured_data_points)
        new_data_points.insert(index, (input_, output))
        self.measured_data_points = tuple(new_data_points)

    def evaluate(
        self,
        required_time_step: TimeStep,
        prepend: int,
        append: int,
        shot_context: ShotContext,
    ) -> DimensionedSeries[np.float64]:
        input_values = self.input_.evaluate(
            required_time_step,
            prepend,
            append,
            shot_context,
        )

        return apply_piecewise_linear_calibration(
            input_values,
            self.measured_data_points,
            Unit(self.input_units) if self.input_units is not None else dimensionless,
            Unit(self.output_units) if self.output_units is not None else dimensionless,
        )


def apply_piecewise_linear_calibration(
    values: DimensionedSeries[np.floating],
    calibration_points: Sequence[tuple[float, float]],
    input_point_units: Unit,
    output_point_units: Unit,
) -> DimensionedSeries[np.float64]:
    """Apply a piecewise linear calibration to a sequencer instruction.

    Args:
        values: The instruction to apply the calibration to.
        calibration_points: A sequence of (input, output) tuples that define the
            points to interpolate between.
            The input must be expressed in input_point_units.
            The output must be expressed in output_point_units.
            The points will be sorted by input value before applying the interpolation.
        input_point_units: The units of the input points of the calibration.
        output_point_units: The units of the output points of the calibration.

    Returns:
        A new series of values where each point is obtained by linearly interpolating
        between calibration points.
        This new series of values is expressed in base units corresponding to
        output_point_units.

    Raises:
        InvalidDimensionalityError: If the units of the input points of the calibration
            are not compatible with the units of the values to map.
    """

    input_points = Quantity(
        [x for x, _ in calibration_points], input_point_units
    ).to_base_units()
    output_points = Quantity(
        [y for _, y in calibration_points], output_point_units
    ).to_base_units()

    if input_points.units != values.units:
        raise InvalidDimensionalityError(
            f"Can't apply calibration with units {input_points.units} to "
            f"instruction with units {values.units}"
        )

    calibration = DimensionlessCalibration(
        list(zip(input_points.magnitude, output_points.magnitude, strict=True))
    )
    return DimensionedSeries(
        calibration.apply(values.values.as_type(np.dtype(np.float64))),
        output_points.units,
    )


class DimensionlessCalibration:
    def __init__(self, calibration_points: Sequence[tuple[float, float]]):
        if len(calibration_points) < 2:
            raise ValueError("Calibration must have at least 2 data points")
        input_points = [x for x, _ in calibration_points]
        output_points = [y for _, y in calibration_points]
        sorted_points = sorted(
            zip(input_points, output_points, strict=True), key=lambda x: x[0]
        )
        sorted_input_points = [x for x, _ in sorted_points]
        sorted_output_points = [y for _, y in sorted_points]
        # We add new flat segments before and after the calibration points to ensure
        # that the calibration is defined for all input values.
        self._input_points = np.array([-np.inf] + sorted_input_points + [+np.inf])
        assert np.all(np.diff(self._input_points) >= 0)
        self._output_points = np.array(
            [sorted_output_points[0]]
            + sorted_output_points
            + [sorted_output_points[-1]]
        )

    @property
    def input_points(self):
        return self._input_points[1:-1]

    @property
    def output_points(self):
        return self._output_points[1:-1]

    def __repr__(self):
        points = ", ".join(
            f"({x}, {y})"
            for x, y in zip(self.input_points, self.output_points, strict=True)
        )
        return f"Calibration({points})"

    def apply(
        self, instruction: TimedInstruction[np.float64]
    ) -> TimedInstruction[np.float64]:
        # We raise errors in pathological cases when the calibration is not
        # well-defined.
        with np.errstate(
            # Avoids ambiguous situation with 2 points at the same x coordinate, because
            # we don't know what to output for a vertical line.
            divide="raise",
            # Avoid situations with 2 x points very close that cause and infinite slope.
            over="raise",
            # Avoids situations with 2 x points are equal and 2 y points are equal,
            # which cause a 0/0 division.
            invalid="raise",
        ):
            np.diff(self._output_points) / np.diff(
                self._input_points
            )  # pyright: ignore[reportUnusedExpression]
        return self._apply_without_checks(instruction)

    @functools.singledispatchmethod
    def _apply_without_checks(
        self, instruction: TimedInstruction[np.float64]
    ) -> TimedInstruction[np.float64]:
        raise NotImplementedError(
            f"Don't know how to apply calibration to instruction of type "
            f"{type(instruction)}"
        )

    @_apply_without_checks.register
    def _apply_calibration_pattern(self, pattern: Pattern) -> Pattern[np.float64]:
        result = self._apply_explicit(pattern.array)
        return Pattern.create_without_copy(result)

    def _apply_explicit(self, value):
        result = np.interp(
            x=value,
            xp=self._input_points,
            fp=self._output_points,
        )
        return result

    @_apply_without_checks.register
    def _apply_calibration_concatenation(
        self, concatenation: Concatenated
    ) -> TimedInstruction[np.float64]:
        return concatenate(
            *(
                self._apply_without_checks(instruction)
                for instruction in concatenation.instructions
            )
        )

    @_apply_without_checks.register
    def _apply_calibration_repetition(
        self, repetition: Repeated
    ) -> TimedInstruction[np.float64]:
        return repetition.repetitions * self._apply_without_checks(
            repetition.instruction
        )

    @_apply_without_checks.register
    def _apply_calibration_ramp(self, r: Ramp) -> TimedInstruction[np.float64]:
        # Ramp maps t -> x(t) = a + (b - a) * t / length
        # Calibration maps x -> y in a piecewise linear way
        # We want to map t -> y(t)
        length = len(r)
        a = r.start
        b = r.stop

        if a == b:
            return Pattern([self._apply_explicit(a)]) * length

        def map_x_segment_to_t(x_0, x_1) -> tuple[float, float]:
            if b > a:
                return length * (x_0 - a) / (b - a), length * (x_1 - a) / (b - a)
            else:
                return length * (x_1 - a) / (b - a), length * (x_0 - a) / (b - a)

        # Find the segments in time over which the output y(t) is linear.
        time_segments = []
        for x0, x1 in pairwise(self._input_points):
            time_segments.append(map_x_segment_to_t(x0, x1))

        if b < a:
            time_segments.reverse()

        sections = []
        for lower, higher in time_segments:
            lower = min(max(lower, 0), length)
            higher = min(max(higher, 0), length)
            i_min = math.ceil(lower)
            i_max = math.ceil(higher)
            sections.append((i_min, i_max))

        sub_ramps = []
        for i_min, i_max in sections:
            if i_max == i_min:
                continue
            in_0 = evaluate_ramp(r, i_min)
            y_0 = self._apply_explicit(in_0)
            if i_max == i_min + 1:
                sub_ramps.append(Pattern([y_0]))
            else:
                in_1 = evaluate_ramp(r, i_max - 1)
                y_1 = self._apply_explicit(in_1)
                length = i_max - i_min
                sub_ramp = create_ramp(
                    y_0, y_0 + length * (y_1 - y_0) / (length - 1), i_max - i_min
                )
                sub_ramps.append(sub_ramp)
        return concatenate(*sub_ramps)


def evaluate_ramp(r: Ramp, t) -> float:
    return r.start + (r.stop - r.start) * t / len(r)



from ._configuration import (
    DeviceConfiguration,
    DeviceServerName,
    DeviceConfigType,
)
from ._converter import get_converter, get_structure_hook, get_unstructure_hook
from ._parameter import DeviceParameter

__all__ = [
    "DeviceConfiguration",
    "DeviceParameter",
    "DeviceServerName",
    "DeviceConfigType",
    "get_converter",
    "get_structure_hook",
    "get_unstructure_hook",
]



from __future__ import annotations

from collections.abc import Mapping
from typing import Optional, Any

import attrs
import numpy as np

import caqtus.formatter as fmt
from caqtus.device import DeviceName
from caqtus.shot_compilation import ShotContext
from caqtus.shot_compilation.lane_compilation import DimensionedSeries
from caqtus.shot_compilation.timed_instructions import (
    TimedInstruction,
    Pattern,
)
from caqtus.types.recoverable_exceptions import InvalidValueError, RecoverableException
from caqtus.types.units import dimensionless
from caqtus.types.variable_name import DottedVariableName
from ._trigger_compiler import TriggerableDeviceCompiler
from ..channel_output import ChannelOutput
from ...timing import TimeStep, number_time_steps


@attrs.define
class DeviceTrigger(ChannelOutput):
    """Indicates that the output should be a trigger for a given device.

    Attributes:
        device_name: The name of the device to generate a trigger for.
        default: If the device is not used in the sequence, fallback to this.
    """

    device_name: DeviceName = attrs.field(
        converter=lambda x: DeviceName(str(x)),
        on_setattr=attrs.setters.convert,
    )
    default: Optional[ChannelOutput] = attrs.field(
        default=None,
        validator=attrs.validators.optional(
            attrs.validators.instance_of(ChannelOutput)
        ),
        on_setattr=attrs.setters.validate,
    )

    def __str__(self):
        return f"trig({self.device_name})"

    def evaluate(
        self,
        required_time_step: TimeStep,
        prepend: int,
        append: int,
        shot_context: ShotContext,
    ):
        target_device = self.device_name
        try:
            target_device_compiler = shot_context.get_device_compiler(target_device)
        except KeyError:
            if self.default is not None:
                evaluated_default = self.default.evaluate(
                    required_time_step,
                    prepend,
                    append,
                    shot_context,
                )
                if evaluated_default.units != dimensionless:
                    raise InvalidValueError(
                        f"Default value for trigger for {fmt.device(target_device)} "
                        f"must be dimensionless, got "
                        f"{fmt.unit(evaluated_default.units)}"
                    ) from None
                default_dtype = evaluated_default.values.dtype
                if default_dtype != np.bool_:
                    raise InvalidValueError(
                        f"Default value for trigger for {fmt.device(target_device)} "
                        f"must be boolean, got {default_dtype}"
                    ) from None
                return evaluated_default
            else:
                raise InvalidValueError(
                    f"There is no {fmt.device(target_device)} to generate trigger for"
                ) from None

        if not isinstance(target_device_compiler, TriggerableDeviceCompiler):
            raise DeviceNotTriggerableError(
                f"{fmt.device(target_device)} can't be triggered"
            )

        trigger_values = target_device_compiler.compute_trigger(
            required_time_step, shot_context
        )

        # We check that the values returned by the target device compiler are valid.
        # If they are not, this a programming error, and it will cause an error further
        # down the line.
        # It means the target device compiler doesn't satisfy the trigger interface, so
        # we want to report it early.
        if not isinstance(trigger_values, TimedInstruction):
            raise TypeError(f"Expected {TimedInstruction}, got {type(trigger_values)}")

        if not np.issubdtype(trigger_values.dtype, np.bool_):
            raise TypeError(f"Expected boolean trigger, got {trigger_values.dtype}")

        length = number_time_steps(shot_context.get_shot_duration(), required_time_step)

        if len(trigger_values) != length:
            raise ValueError(
                f"Expected an instruction with length {length}, got "
                f"{len(trigger_values)}"
            )

        return DimensionedSeries(
            prepend * Pattern([False]) + trigger_values + append * Pattern([False]),
            units=dimensionless,
        )

    def evaluate_max_advance_and_delay(
        self,
        time_step: TimeStep,
        variables: Mapping[DottedVariableName, Any],
    ) -> tuple[int, int]:
        return 0, 0


class DeviceNotTriggerableError(RecoverableException):
    pass



"""Provides a widget for editing time lanes.

The main class is :class:`TimeLanesEditor` which is a widget that displays multiple
time lanes, along with the time steps, in a table-like view.

To know how to display and edit a specific type of lane, the editor uses
the :class:`extension.CondetrolLaneExtensionProtocol` to know how to build a
:class:`TimeLaneModel` and a :class:`TimeLaneDelegate` for each lane.
"""

from . import digital_lane_editor, analog_lane_editor, camera_lane_editor, extension
from ._delegate import TimeLaneDelegate
from ._time_lanes_editor import TimeLanesEditor
from ._time_lane_model import TimeLaneModel

__all__ = [
    "TimeLanesEditor",
    "digital_lane_editor",
    "analog_lane_editor",
    "camera_lane_editor",
    "TimeLaneModel",
    "TimeLaneDelegate",
    "extension",
]



from datetime import datetime
from datetime import timezone
from typing import TYPE_CHECKING, assert_never, Optional

import sqlalchemy.orm
from attr import frozen
from sqlalchemy import select
from sqlalchemy.orm import Session

from caqtus.utils.result import (
    Success,
    Failure,
    is_failure,
    is_failure_type,
    unwrap,
)
from ._path_table import SQLSequencePath
from .._path import PureSequencePath
from .._path_hierarchy import (
    PathNotFoundError,
    PathIsRootError,
    PathHierarchy,
    PathExistsError,
    RecursivePathMoveError,
)
from .._sequence_collection import PathIsSequenceError, SequenceRunningError

if TYPE_CHECKING:
    from ._experiment_session import SQLExperimentSession


@frozen
class SQLPathHierarchy(PathHierarchy):
    parent_session: "SQLExperimentSession"

    def does_path_exists(self, path: PureSequencePath) -> bool:
        return _does_path_exists(self._get_sql_session(), path)

    def create_path(
        self, path: PureSequencePath
    ) -> Success[list[PureSequencePath]] | Failure[PathIsSequenceError]:
        paths_to_create: list[PureSequencePath] = []
        current = path
        sequence_collection = self.parent_session.sequences
        while parent := current.parent:
            is_sequence_result = sequence_collection.is_sequence(current)
            match is_sequence_result:
                case Success(is_sequence):
                    if is_sequence:
                        return Failure(
                            PathIsSequenceError(
                                f"Cannot create path {path} because {current} is "
                                f"already a sequence"
                            )
                        )
                case Failure(PathNotFoundError()):
                    paths_to_create.append(current)
                case _:
                    assert_never(is_sequence_result)
            current = parent

        session = self._get_sql_session()
        created_paths = []
        for path_to_create in reversed(paths_to_create):
            assert path_to_create.parent is not None
            parent_model_result = _query_path_model(session, path_to_create.parent)
            if isinstance(parent_model_result, Failure):
                assert is_failure_type(parent_model_result, PathIsRootError)
                parent_model = None
            else:
                parent_model = parent_model_result.value

            new_path = SQLSequencePath(
                path=str(path_to_create),
                parent=parent_model,
                creation_date=datetime.now(tz=timezone.utc).replace(tzinfo=None),
            )
            session.add(new_path)
            created_paths.append(path_to_create)
        return Success(created_paths)

    def get_children(
        self, path: PureSequencePath
    ) -> (
        Success[set[PureSequencePath]]
        | Failure[PathNotFoundError]
        | Failure[PathIsSequenceError]
    ):
        return _get_children(self._get_sql_session(), path)

    def delete_path(
        self, path: PureSequencePath, delete_sequences: bool = False
    ) -> (
        Success[None]
        | Failure[PathNotFoundError]
        | Failure[PathIsSequenceError]
        | Failure[PathIsRootError]
    ):

        session = self._get_sql_session()

        if not delete_sequences:
            sequence_collection = self.parent_session.sequences
            contained_sequence_result = sequence_collection.get_contained_sequences(
                path
            )
            if is_failure(contained_sequence_result):
                return contained_sequence_result
            if contained_sequence_result.value:
                return Failure(
                    PathIsSequenceError(
                        f"Cannot delete a path that contains sequences: "
                        f"{contained_sequence_result.value}"
                    )
                )

        path_model_result = _query_path_model(session, path)
        if is_failure_type(path_model_result, PathIsRootError):
            return path_model_result
        assert not is_failure_type(path_model_result, PathNotFoundError)
        session.delete(path_model_result.value)
        return Success(None)

    def get_all_paths(self) -> set[PureSequencePath]:
        query = select(SQLSequencePath)
        result = self._get_sql_session().execute(query)
        return {PureSequencePath(path.path) for path in result.scalars()}

    def update_creation_date(self, path: PureSequencePath, date: datetime) -> None:
        if path.is_root():
            raise PathIsRootError(path)

        sql_path = unwrap(self._query_path_model(path))
        sql_path.creation_date = date.astimezone(timezone.utc).replace(tzinfo=None)

    def _query_path_model(
        self, path: PureSequencePath
    ) -> (
        Success[SQLSequencePath] | Failure[PathNotFoundError] | Failure[PathIsRootError]
    ):
        return _query_path_model(self._get_sql_session(), path)

    def _get_sql_session(self) -> sqlalchemy.orm.Session:
        # noinspection PyProtectedMember
        return self.parent_session._get_sql_session()

    def get_path_creation_date(
        self, path: PureSequencePath
    ) -> Success[datetime] | Failure[PathNotFoundError] | Failure[PathIsRootError]:
        return _get_path_creation_date(self._get_sql_session(), path)

    def move(
        self, source: PureSequencePath, destination: PureSequencePath
    ) -> (
        Success[None]
        | Failure[PathNotFoundError]
        | Failure[PathExistsError]
        | Failure[PathIsSequenceError]
        | Failure[RecursivePathMoveError]
        | Failure[SequenceRunningError]
    ):
        if destination.is_descendant_of(source) or source == destination:
            return Failure(
                RecursivePathMoveError(f"Cannot move {source} into {destination}")
            )
        session = self._get_sql_session()
        source_path_result = _query_path_model(session, source)
        if isinstance(source_path_result, Failure):
            # We can't have the PathIsRootError here because it is prevented by
            # ensuring that the destination can't be a descendant of the source.
            assert not is_failure_type(source_path_result, PathIsRootError)
            return source_path_result
        source_model = source_path_result.result()

        running_sequences = unwrap(
            self.parent_session.sequences.get_contained_running_sequences(source)
        )
        if running_sequences:
            return Failure(
                SequenceRunningError(
                    f"Cannot move {source} because it contains running sequences: "
                    f"{running_sequences}"
                )
            )

        if self.does_path_exists(destination):
            return Failure(PathExistsError(destination))

        # destination doesn't exist, so it can't be the root path, and thus it has a
        # parent.
        assert destination.parent is not None

        created_paths_result = self.create_path(destination.parent)
        if isinstance(created_paths_result, Failure):
            return created_paths_result
        if destination.parent.is_root():
            destination_parent_model = None
        else:
            destination_parent_model = unwrap(
                _query_path_model(session, destination.parent)
            )
        source_model.parent_id = (
            destination_parent_model.id_ if destination_parent_model else None
        )
        source_model.path = str(destination)

        descendants = self._get_sql_session().query(
            self.descendants_query(source_model.id_)
        )
        for child in descendants.all():
            child_path = PureSequencePath(str(child.path))
            new_path = PureSequencePath.from_parts(
                destination.parts + child_path.parts[len(source.parts) :]
            )
            child.path = str(new_path)

        return Success(None)

    def check_valid(self) -> None:
        """Check that the hierarchy is valid."""

        # We check that the name of the paths are consistent with the links between
        # them.
        for child in unwrap(self.get_children(PureSequencePath.root())):
            self._check_valid(unwrap(self._query_path_model(child)))

    def _check_valid(self, path: SQLSequencePath) -> None:
        current_path = PureSequencePath(str(path.path))
        for child in path.children:
            child_path = PureSequencePath(str(child.path))
            if child_path.parent != current_path:  # pragma: no cover
                raise AssertionError("Invalid path hierarchy")
            self._check_valid(child)

    @staticmethod
    def descendants_query(
        ancestor_id: Optional[int],
    ) -> type[SQLSequencePath]:
        """Return an expression for querying the descendants of a path.

        Args:
            ancestor_id: The id of the ancestor path.
                If None, the root path is used.
                The ancestor is not included in the descendants.

        Returns:
            A expression representing the descendants of the ancestor.
        """

        desc_cte = (
            select(SQLSequencePath)
            .where(SQLSequencePath.parent_id == ancestor_id)
            .cte(recursive=True)
        )
        desc_cte = desc_cte.union(
            select(SQLSequencePath).filter(SQLSequencePath.parent_id == desc_cte.c.id)
        )
        query = sqlalchemy.orm.aliased(SQLSequencePath, desc_cte)

        return query


def _does_path_exists(session: Session, path: PureSequencePath) -> bool:
    if path.is_root():
        return True
    result = _query_path_model(session, path)
    return isinstance(result, Success)


def _get_children(
    session: Session, path: PureSequencePath
) -> (
    Success[set[PureSequencePath]]
    | Failure[PathNotFoundError]
    | Failure[PathIsSequenceError]
):
    query_result = _query_path_model(session, path)
    if isinstance(query_result, Success):
        path_sql = unwrap(query_result)
        if path_sql.sequence:
            return Failure(PathIsSequenceError(str(path)))
        else:
            children = path_sql.children
    elif isinstance(query_result, Failure):
        if is_failure_type(query_result, PathIsRootError):
            query_children = select(SQLSequencePath).where(
                SQLSequencePath.parent_id.is_(None)
            )
            children = session.scalars(query_children)
        elif is_failure_type(query_result, PathNotFoundError):
            return query_result
        else:
            assert_never(query_result)
    else:
        assert_never(query_result)

    return Success(set(PureSequencePath(str(child.path)) for child in children))


def _get_path_creation_date(
    session: Session, path: PureSequencePath
) -> Success[datetime] | Failure[PathNotFoundError] | Failure[PathIsRootError]:
    return _query_path_model(session, path).map(
        lambda x: x.creation_date.replace(tzinfo=timezone.utc)
    )


def _query_path_model(
    session: Session, path: PureSequencePath
) -> Success[SQLSequencePath] | Failure[PathNotFoundError] | Failure[PathIsRootError]:
    if path.is_root():
        return Failure(PathIsRootError(path))
    stmt = select(SQLSequencePath).where(SQLSequencePath.path == str(path))
    result = session.execute(stmt)
    if found := result.scalar():
        return Success(found)
    else:
        return Failure(PathNotFoundError(f'Path "{path}" does not exists'))



from __future__ import annotations

from collections.abc import Callable
from typing import Generic

import attrs

from caqtus.device import DeviceConfiguration, Device, DeviceController
from caqtus.device.configuration import DeviceConfigType as DeviceConfigT
from caqtus.device.remote import DeviceProxy
from caqtus.gui.condetrol.device_configuration_editors import DeviceConfigurationEditor
from caqtus.shot_compilation import DeviceCompiler
from caqtus.utils.serialization import JSON


@attrs.frozen(kw_only=True)
class DeviceExtension(Generic[DeviceConfigT]):
    """Define how to implement a device plugin.

    This class is generic in the :class:`DeviceConfiguration` type.

    Attributes:
        label: A human-readable label for the type of device represented by this
            extension.

            This label describes the type of device and will be displayed to the user
            when they are selecting a device to add to the experiment.

            No two device extensions should have the same label.

        configuration_type: The type of configuration used to store the settings of
            the device.

            The name of the type is used to identify the configuration type in the
            storage.
            It is thus not recommended to change the name of the type after it has been
            used in the application, as the new name will not be recognized as the same
            type.

        configuration_factory: A factory function that returns a new instance of the
            configuration type.

            This function will be called when a new device of this type is added to the
            experiment.
            It should return a default instance of the configuration type.

        configuration_dumper: A function that converts a configuration instance to a
            JSON-serializable format.
            This function will be used to save the configuration.

        configuration_loader: A function that converts a JSON-serializable format to a
            configuration instance.
            This function will be used to load the configuration.
            It is passed the JSON-serializable format saved by the
            `configuration_dumper` function.

        editor_type: A function that creates an editor for the device configuration.

            When the user wants to edit the configuration of a device of this type,
            this function will be called to create an editor for the configuration.
            The function is passed as argument the configuration instance to edit.

            Once the user has finished editing the configuration, the method
            :meth:`~caqtus.gui.condetrol.device_configuration_editors.DeviceConfigurationEditor.get_configuration`
            of the editor is called and the result is saved.

        device_type: A callable that returns a new device instance.

            This function will be called when a device associated with this extension
            needs to be created.
            The arguments passed are obtained by calling the method
            :meth:`caqtus.device.compiler.DeviceCompiler.compile_initialization_parameters`
            of the compiler associated with the device.

            This function needs to be pickleable as it will be sent to a remote process.

        compiler_type: The type of compiler used to compile the parameters of the
            device.

            When the sequence is launched, a compiler will be created for each device
            configuration. The compiler
            :meth:`~caqtus.shot_compilation.DeviceCompiler.__init__` method is
            called with the name of the device and the context of the sequence.
            If the initialization method raises :class:`DeviceNotUsedException`, no
            further actions will be taken for this device, and it will be ignored for
            the sequence.

            The method
            :meth:`~caqtus.shot_compilation.DeviceCompiler.compile_initialization_parameters`
            is called at the beginning of the sequence to get the parameters to
            pass to the :attr:`device_type` function to create the device.

            The method
            :meth:`~caqtus.shot_compilation.DeviceCompiler.compile_shot_parameters` is
            then called for each shot to get the parameters to pass to the device
            controller for the shot.

            Warning:
                The compilation of each shot happens in a new process using each time a
                new copy of the compiler.
                This has some consequences:

                - The compiler should be pickleable.
                - If the compiler is modified during the compilation of a shot, the
                  changes will not be reflected in the next shot.
                - :func:`print` statements in the compiler will not be displayed in the
                  console.

    """

    label: str = attrs.field(converter=str)

    configuration_type: type[DeviceConfigT] = attrs.field()
    configuration_factory: Callable[[], DeviceConfigT] = attrs.field()
    configuration_dumper: Callable[[DeviceConfigT], JSON] = attrs.field()
    configuration_loader: Callable[[JSON], DeviceConfigT] = attrs.field()

    editor_type: Callable[[DeviceConfigT], DeviceConfigurationEditor[DeviceConfigT]] = (
        attrs.field()
    )

    device_type: Callable[..., Device] = attrs.field()
    compiler_type: type[DeviceCompiler] = attrs.field()
    controller_type: type[DeviceController] = attrs.field()
    proxy_type: type[DeviceProxy] = attrs.field()

    @configuration_type.validator  # type: ignore
    def _validate_configuration_type(self, _, value):
        if not issubclass(value, DeviceConfiguration):
            raise ValueError(f"Invalid configuration type: {value}.")

    @compiler_type.validator  # type: ignore
    def _validate_compiler_type(self, _, value):
        if not issubclass(value, DeviceCompiler):
            raise ValueError(f"Invalid compiler type: {value}.")

    @controller_type.validator  # type: ignore
    def _validate_controller_type(self, _, value):
        if not issubclass(value, DeviceController):
            raise ValueError(f"Invalid controller type: {value}.")

    @proxy_type.validator  # type: ignore
    def _validate_proxy_type(self, _, value):
        if not issubclass(value, DeviceProxy):
            raise ValueError(f"Invalid proxy type: {value}.")



from __future__ import annotations

import abc
from collections.abc import Mapping
from typing import Protocol, Any, runtime_checkable

from caqtus.device import DeviceName
from .compilation_contexts import SequenceContext, ShotContext


@runtime_checkable
class DeviceCompiler(Protocol):
    """Defines the interface for a device compiler.

    Each device compiler is responsible to evaluate the parameters to apply to a device.

    Its role is to translate high-level and user-friendly parameters into concrete
    low-level device parameters that can be used to program the device.

    The compiler is initialized at the beginning of a sequence.
    Its method :meth:`compile_initialization_parameters` is then called once to
    obtain the parameters to pass to the device constructor.

    The for each shot, the method :meth:`compile_shot_parameters` is called to pass to
    the device controller for this shot.

    If it is necessary to generate a trigger for the device under consideration, the
    device compiler should inherit from
    :class:`caqtus.device.sequencer.compilation.TriggerableDeviceCompiler`.
    """

    @abc.abstractmethod
    def __init__(self, device_name: DeviceName, sequence_context: SequenceContext):
        """Initialize the device compiler.

        Args:
            device_name: The name of the device for which the compiler is being created.
                The device name can be used to retrieve the device configuration in the
                sequence context.
            sequence_context: The context of the sequence being compiled.
                It contains information about the current that can be useful to evaluate
                device parameters.
        Raises:
            DeviceNotUsedException: If the device is not used in the current sequence.
        """

        pass

    @abc.abstractmethod
    def compile_initialization_parameters(self) -> Mapping[str, Any]:
        """Compile the parameters to pass to the device constructor."""

        return {}

    @abc.abstractmethod
    def compile_shot_parameters(self, shot_context: ShotContext) -> Mapping[str, Any]:
        """Compile the parameters to pass to the device controller for a shot.

        Args:
            shot_context: The context of the shot being compiled.
                It contains information about the shot that can be useful to evaluate
                device parameters.
        """

        return {}


class DeviceNotUsedException(Exception):  # noqa: N818
    """Raised when a device is not used in a sequence."""

    pass



from __future__ import annotations

from typing import Any

from typing_extensions import TypeIs

type JSON = JsonDict | JsonList | str | int | float | bool | None
type JsonDict = dict[str, JSON]
type JsonList = list[JSON]


def is_valid_json_dict(data: Any) -> TypeIs[JsonDict]:
    if not isinstance(data, dict):
        return False
    valid_keys = all(isinstance(key, str) for key in data.keys())
    valid_values = all(is_valid_json(value) for value in data.values())
    return valid_keys and valid_values


def is_valid_json_list(data: Any) -> TypeIs[JsonList]:
    if not isinstance(data, list):
        return False
    return all(is_valid_json(value) for value in data)


def is_valid_json(data) -> TypeIs[JSON]:
    if isinstance(data, dict):
        return is_valid_json_dict(data)
    elif isinstance(data, list):
        return is_valid_json_list(data)
    elif data is None:
        return True
    elif isinstance(data, (str, int, float, bool)):
        return True
    else:
        return False



from __future__ import annotations

from collections.abc import Mapping, Iterable
from typing import Any, Optional, overload

import attrs
import numpy as np

from caqtus.types._array import FloatArray1D
from caqtus.types.units import (
    Unit,
    Quantity,
    dimensionless,
    is_scalar_quantity,
)
from caqtus.types.variable_name import DottedVariableName
from ._transformation import (
    Transformation,
    EvaluableOutput,
    evaluable_output_validator,
    OutputValue,
    evaluate,
)


def _data_points_converter(data_points: Iterable[tuple[float, float]]):
    point_to_tuple = [(x, y) for x, y in data_points]
    return tuple(sorted(point_to_tuple))


@attrs.define
class LinearInterpolation(Transformation):
    """Transforms an input value by applying a piecewise linear interpolation.

    This transformation stores a set of measured data points and interpolates the
    output value based on the input value.

    Attributes:
        input_: An operation that can be evaluated to an output value.
            The transformation is applied to this value.
        measured_data_points: A list of measured data points as tuples of input and
            output values.
        input_points_unit: The unit of the input points.
            The result of the input evaluation will be converted to this unit.
        output_points_unit: The unit of the output points.
            The result of the transformation will be converted to this unit.
    """

    input_: EvaluableOutput = attrs.field(
        validator=evaluable_output_validator,
        on_setattr=attrs.setters.validate,
    )
    measured_data_points: tuple[tuple[float, float], ...] = attrs.field(
        converter=_data_points_converter, on_setattr=attrs.setters.convert
    )
    input_points_unit: Optional[str] = attrs.field(
        converter=attrs.converters.optional(str),
        on_setattr=attrs.setters.convert,
    )
    output_points_unit: Optional[str] = attrs.field(
        converter=attrs.converters.optional(str),
        on_setattr=attrs.setters.convert,
    )

    def evaluate(self, variables: Mapping[DottedVariableName, Any]) -> OutputValue:
        evaluated = evaluate(self.input_, variables)
        if isinstance(evaluated, Quantity):
            input_value = evaluated
        else:
            input_value = Quantity(float(evaluated), dimensionless)
        result = interpolate(
            input_value,
            Quantity(
                [point[0] for point in self.measured_data_points],
                (
                    Unit(self.input_points_unit)
                    if self.input_points_unit
                    else dimensionless
                ),
            ),
            Quantity(
                [point[1] for point in self.measured_data_points],
                (
                    Unit(self.output_points_unit)
                    if self.output_points_unit
                    else dimensionless
                ),
            ),
        )
        return result.to_base_units()


@overload
def interpolate[
    InputUnits: Unit, OutputUnits: Unit, L: int
](
    values: Quantity[float, InputUnits],
    input_values: Quantity[FloatArray1D[L], InputUnits],
    output_values: Quantity[FloatArray1D[L], OutputUnits],
) -> Quantity[float, OutputUnits]: ...


@overload
def interpolate[
    InputUnits: Unit, OutputUnits: Unit, L: int
](
    values: Quantity[FloatArray1D, InputUnits],
    input_values: Quantity[FloatArray1D[L], InputUnits],
    output_values: Quantity[FloatArray1D[L], OutputUnits],
) -> Quantity[FloatArray1D, OutputUnits]: ...


def interpolate[
    InputUnits: Unit, OutputUnits: Unit, L: int
](
    values: Quantity[FloatArray1D | float, InputUnits],
    input_values: Quantity[FloatArray1D[L], InputUnits],
    output_values: Quantity[FloatArray1D[L], OutputUnits],
) -> Quantity[FloatArray1D | float, OutputUnits]:
    sorted_points = sorted(
        zip(input_values.magnitude, output_values.magnitude, strict=True),
        key=lambda x: x[0],
    )

    base_input_values = Quantity(
        [point[0] for point in sorted_points], input_values.units
    ).to_base_units()
    base_output_values = Quantity(
        [point[1] for point in sorted_points], output_values.units
    ).to_base_units()

    output_magnitude = np.interp(
        x=values.to_unit(base_input_values.units).magnitude,
        xp=base_input_values.magnitude,
        fp=base_output_values.magnitude,
    )
    if is_scalar_quantity(values):
        return Quantity(float(output_magnitude), base_output_values.units).to_unit(
            output_values.units
        )
    else:
        return Quantity[FloatArray1D](
            output_magnitude, base_output_values.units
        ).to_unit(output_values.units)



from caqtus.gui._common.NodeGraphQt import BaseNode
from caqtus.types.expression import Expression


class ConstantNode(BaseNode):
    __identifier__ = "caqtus.sequencer_node.source"
    NODE_NAME = "Constant"

    def __init__(self):
        super().__init__()
        self.add_output("out", multi_output=False, display_name=False)
        self.add_text_input(
            "Value",
            text="...",
            placeholder_text="value",
            tooltip="The output of this node is held constant during the whole shot.",
        )

    def set_value(self, expression: Expression) -> None:
        self.set_property("Value", str(expression))

    def get_value(self) -> Expression:
        return Expression(str(self.get_property("Value")))



from PySide6.QtWidgets import QSpinBox

from ._value_editor import ValueEditor


class IntegerEditor(ValueEditor[int]):
    """An editor to display an integer.

    Args:
        min_value: The lowest value (inclusive) that can be entered.
        max_value: The largest value (inclusive) that can be entered.
        prefix: A string to be displayed before the number.
        suffix: A string to be displayed after the number.
    """

    def __init__(
        self,
        min_value=0,
        max_value=999,
        prefix: str = "",
        suffix: str = "",
    ) -> None:
        self.spin_box = QSpinBox()
        self.spin_box.setRange(min_value, max_value)
        self._min_value = min_value
        self._max_value = max_value

        if prefix:
            self.spin_box.setPrefix(prefix)
        if suffix:
            self.spin_box.setSuffix(suffix)

    def set_value(self, value: int) -> None:
        if not self._min_value <= value <= self._max_value:
            raise ValueError(
                f"Value {value} is outside the editor range "
                f"[{self._min_value}, {self._max_value}]"
            )
        self.spin_box.setValue(value)

    def read_value(self) -> int:
        return self.spin_box.value()

    def set_editable(self, editable: bool) -> None:
        self.spin_box.setReadOnly(not editable)

    @property
    def widget(self) -> QSpinBox:
        return self.spin_box



from __future__ import annotations

import abc
from collections.abc import Mapping
from typing import TypeGuard

from ..parameter import Parameter, ParameterSchema
from ..variable_name import DottedVariableName


class IterationConfiguration(abc.ABC):
    """Defines how parameters should be iterated over.

    This is an abstract base class that defines the interface for iterations of shots
    during a sequence.
    It is meant to be subclassed to define different types of iterations.
    """

    @abc.abstractmethod
    def expected_number_shots(self) -> int | Unknown:
        """Return the expected number of shots defined by this iteration.

        If the number of shots can be statically determined ahead of time, this method
        should return that number.
        If the number of shots cannot be determined ahead of time, this method should
        return unknown.
        In doubt, the method must return unknown and not a possibly wrong guess.
        """

        raise NotImplementedError

    @abc.abstractmethod
    def get_parameter_names(self) -> set[DottedVariableName]:
        """Return the names of the parameters that are iterated over.

        This method must return the name of the parameters whose values are changed
        during the iteration.
        The iteration must set the values of all these parameters before each shot.
        """

        raise NotImplementedError

    @abc.abstractmethod
    def get_parameter_schema(
        self, initial_parameters: Mapping[DottedVariableName, Parameter]
    ) -> ParameterSchema:
        """Compute the schema of the parameters that are iterated over.

        Args:
            initial_parameters: The values of the parameters that are defined before the
                iteration starts.
                Unless the iteration overwrites the values of these parameters, they
                are considered constant during the iteration.

        Returns:
            The schema of the parameters that are iterated over.

            Implementations of this method should infer the types of the parameters
            correctly as the consumers of this method rely on the correctness of the
            inferred types.
        """

        raise NotImplementedError


class Unknown:
    """Instances of this class represent an unknown positive integer.

    It can be added or multiplied with a positive integer.
    It is absorbent for addition and for multiplication with a non-zero operand.
    Multiplication by zero return zero.
    """

    def __add__(self, other):
        match other:
            case int(x) if x >= 0:
                return self
            case Unknown():
                return self
            case _:
                return NotImplemented

    def __radd__(self, other):
        return self.__add__(other)

    def __mul__(self, other):
        match other:
            case 0:
                return 0
            case Unknown():
                return self
            case int(x) if x > 0:
                return self
            case _:
                return NotImplemented

    def __rmul__(self, other):
        return self.__mul__(other)

    def __repr__(self):
        return "Unknown()"

    def __str__(self):
        return "unknown"


def is_unknown(value) -> TypeGuard[Unknown]:
    return isinstance(value, Unknown)



"""Allow to manage the high-level control of experiments."""

from . import sequence_execution
from .manager import ExperimentManager, Procedure

__all__ = [
    "ExperimentManager",
    "Procedure",
    "sequence_execution",
]



from __future__ import annotations

import importlib.resources
from collections.abc import Sequence
from typing import NewType, SupportsFloat
from typing import overload, Generic

import numpy as np
import pint._typing
import pint.facets
import pint.facets.nonmultiplicative.objects
import pint.facets.numpy.quantity
import pint.facets.numpy.unit
from typing_extensions import TypeIs, TypeVar

from caqtus.types.recoverable_exceptions import InvalidValueError
from .._array import FloatArray

UnitLike = pint._typing.UnitLike


class Unit(
    pint.facets.SystemRegistry.Unit,
    pint.facets.numpy.unit.NumpyUnit,
    pint.facets.nonmultiplicative.objects.NonMultiplicativeUnit,
    pint.facets.plain.PlainUnit,
):
    def to_base(self) -> BaseUnit:
        """Convert the unit to base units."""

        return BaseUnit(Quantity(1.0, self).to_base_units().units)

    def __pow__(self, power) -> Unit:
        result = super().__pow__(power)
        assert isinstance(result, Unit)
        return result

    def __repr__(self) -> str:
        return f"Unit(\"{format(self, '~')}\")"


BaseUnit = NewType("BaseUnit", Unit)
"""A type that represents a unit expressed in base SI units."""


type Magnitude = float | FloatArray

M = TypeVar("M", bound=Magnitude, covariant=True, default=Magnitude)
U = TypeVar("U", bound=Unit, covariant=True, default=Unit)
V = TypeVar("V", bound=Unit, covariant=True, default=Unit)
A = TypeVar("A", bound=FloatArray, covariant=True, default=FloatArray)


class Quantity(
    pint.facets.system.objects.SystemQuantity[M],  # type: ignore[reportInvalidTypeArguments]
    pint.facets.numpy.quantity.NumpyQuantity[M],  # type: ignore[reportInvalidTypeArguments]
    pint.facets.nonmultiplicative.objects.NonMultiplicativeQuantity[M],  # type: ignore[reportInvalidTypeArguments]
    pint.facets.plain.PlainQuantity[M],  # type: ignore[reportInvalidTypeArguments]
    Generic[M, U],
):
    @overload
    def __new__(cls, value: int, units: V) -> Quantity[float, V]: ...

    @overload
    def __new__(cls, value: M, units: V) -> Quantity[M, V]: ...

    @overload
    def __new__(
        cls, value: Sequence[SupportsFloat], units: V
    ) -> Quantity[FloatArray, V]: ...

    def __new__(cls, value: int | Magnitude | Sequence[SupportsFloat], units: Unit):
        if isinstance(value, int):
            return super().__new__(
                cls,
                float(value),  # type: ignore[reportArgumentType]
                units,
            )
        return super().__new__(cls, value, units)  # type: ignore[reportArgumentType]

    @property
    def units(self) -> U:
        u = super().units
        return u  # type: ignore[reportReturnType]

    def to_base_units(self) -> Quantity[M, "BaseUnit"]:
        result = super().to_base_units()
        assert isinstance(result, Quantity)
        return result

    def to_unit(self, unit: V) -> Quantity[M, V]:
        """

        Raises:
            DimensionalityError: If the units are not compatible.
        """
        result = super().to(unit)
        assert isinstance(result, Quantity)
        return result

    @property
    def magnitude(self) -> M:
        mag = super().magnitude
        if isinstance(mag, np.ndarray):
            return mag.astype(float)  # type: ignore[reportReturnType]
        else:
            return float(mag)  # type: ignore[reportReturnType]

    def __str__(self):
        return format(self, "~")

    def __repr__(self) -> str:
        return f"Quantity({self.magnitude}, {self.units!r})"


def is_quantity(value) -> TypeIs[Quantity]:
    """Returns True if the value is a quantity, False otherwise."""

    return isinstance(value, Quantity)


def is_scalar_quantity(value) -> TypeIs[Quantity[float]]:
    """Returns True if the value is a scalar quantity, False otherwise."""

    return is_quantity(value) and isinstance(value.magnitude, float)


class UnitRegistry(pint.UnitRegistry):
    Quantity = Quantity  # type: ignore[reportAssignmentType]
    Unit = Unit  # type: ignore[reportAssignmentType]


units_definition_file = importlib.resources.files("caqtus.types.units").joinpath(
    "units_definition.txt"
)

ureg = UnitRegistry(
    str(units_definition_file),
    autoconvert_offset_to_baseunit=True,
    cache_folder=":auto:",
)
unit_registry = ureg
pint.set_application_registry(unit_registry)


UndefinedUnitError = pint.UndefinedUnitError

DimensionalityError = pint.DimensionalityError
Dimensionless = NewType("Dimensionless", BaseUnit)
dimensionless = Dimensionless(BaseUnit(Unit("dimensionless")))

TIME_UNITS = {"s", "ms", "µs", "us", "ns"}

FREQUENCY_UNITS = {
    "mHz",
    "Hz",
    "kHz",
    "MHz",
    "GHz",
    "THz",
}

POWER_UNITS = {
    "W",
    "mW",
    "µW",
    "uW",
    "nW",
    "dBm",
}

DIMENSIONLESS_UNITS = {"dB", "percent", "%"}

CURRENT_UNITS = {"A", "mA", "µA", "uA", "nA"}

VOLTAGE_UNITS = {"V", "mV", "µV", "uV", "nV"}

DISTANCE_UNITS = {"m", "mm", "µm", "um", "nm"}

ANGLE_UNITS = {"deg", "°", "rad"}

UNITS = (
    TIME_UNITS
    | FREQUENCY_UNITS
    | POWER_UNITS
    | DIMENSIONLESS_UNITS
    | CURRENT_UNITS
    | VOLTAGE_UNITS
    | DISTANCE_UNITS
    | ANGLE_UNITS
)


class InvalidDimensionalityError(InvalidValueError):
    """Raised when a value has an invalid dimensionality.

    This error is raised when a value has an invalid dimensionality and the user
    should fix it.
    """

    pass



from ._evaluate_scalar_expression import evaluate_scalar_expression
from ._exceptions import UndefinedParameterError
from ._time_dependent_expression import evaluate_time_dependent_digital_expression

__all__ = [
    "evaluate_scalar_expression",
    "UndefinedParameterError",
    "evaluate_time_dependent_digital_expression",
]



from typing import NewType

import numpy as np

type FloatArray[Shape: tuple[int, ...]] = np.ndarray[Shape, np.dtype[np.floating]]
type FloatArray1D[Length: int] = FloatArray[tuple[Length]]

SortedFloatArray = NewType("SortedFloatArray", FloatArray1D)



"""This module defines what can be output by a sequencer channel.

Classes that inherit from :class:`ChannelOutput` declare what should be output by a
channel.
They can be evaluated to a :class:`EvaluatedOutput` object, which contains the sequence
of values to output and their units.

A channel output can be dependent on another channel output.
This allows the user to build complex evaluation pipelines for what should be output
by a sequencer channel.
"""

from . import timing
from ._calibrated_analog_mapping import CalibratedAnalogMapping, TimeIndependentMapping
from ._channel_sources import (
    LaneValues,
    Constant,
    DeviceTrigger,
    ValueSource,
    is_value_source,
)
from .channel_output import ChannelOutput

__all__ = [
    "ChannelOutput",
    "LaneValues",
    "DeviceTrigger",
    "Constant",
    "ValueSource",
    "is_value_source",
    "CalibratedAnalogMapping",
    "TimeIndependentMapping",
    "timing",
]



import enum
import typing
from collections.abc import Callable

import attrs

from .._value_editor import ValueEditor

# TODO: Use PEP 747 if accepted
type TypeExpr[T] = typing.Any

type EditorFactory[T] = Callable[[], ValueEditor[T]]
"""A function that can be used to create an editor for a value."""


class EditorBuilder:
    """Construct widgets to edit values of given types.

    Editors must be registered with the builder using the method :meth:`register_editor`
    before the method :meth:`build_editor` is called.
    """

    def __init__(self) -> None:
        self._editor_factories: dict[TypeExpr, EditorFactory] = {}

    def register_editor[T](self, type_: TypeExpr[T], factory: EditorFactory) -> None:
        """Specify an editor to use when encountering a given type.

        When the method :meth:`build_editor` is called with this type, this editor
        will be used.

        Warning:
            If an editor is already registered for this type, it will be overwritten.
        """

        self._editor_factories[type_] = factory

    def build_editor(self, type_: TypeExpr) -> EditorFactory:
        """Construct a gui class to edit value of a given type.

        This method dispatches the type passed as argument to the editor registered
        for this type.

        If the type is `typing.Annotated[T, ...]`, the editor for `T` will be built.

        If the type is not registered, but is an attrs class, an editor will be built
        for it by calling :func:`build_attrs_class_editor`.

        If the type is not registered but is a `typing.Literal`, an editor will be built
        for it by calling :func:`build_literal_editor`.

        Raises:
            TypeNotRegisteredError: If no editor is registered to handle the given type,
                and it is not possible to infer an editor from the type.
        """

        origin = typing.get_origin(type_)

        if origin is typing.Annotated:  # typing.Annotated[T, ...]
            try:
                return self.build_editor(typing.get_args(type_)[0])
            except Exception as e:
                e.add_note(f"The error occurred while processing the type {type_}")
                raise e from e.__cause__

        try:
            return self._editor_factories[type_]
        except KeyError:
            if origin == typing.Literal:
                from ._literal import build_literal_editor

                return build_literal_editor(*typing.get_args(type_))
            elif issubclass(type_, enum.Enum):
                from ._enum import EnumEditor, generate_enum_editor

                return generate_enum_editor(type_)
            elif attrs.has(type_):
                from ._attrs import build_attrs_class_editor

                return build_attrs_class_editor(type_, self)
            else:
                raise TypeNotRegisteredError(
                    f"No editor is registered to handle {type_}"
                ) from None


class EditorBuildingError(Exception):
    pass


class TypeNotRegisteredError(EditorBuildingError):
    pass



from .iteration_configuration import IterationConfiguration, Unknown, is_unknown
from .steps_configurations import (
    StepsConfiguration,
    Step,
    ExecuteShot,
    VariableDeclaration,
    LinspaceLoop,
    ArangeLoop,
    ContainsSubSteps,
)

__all__ = [
    "IterationConfiguration",
    "Step",
    "StepsConfiguration",
    "ExecuteShot",
    "VariableDeclaration",
    "LinspaceLoop",
    "ArangeLoop",
    "ContainsSubSteps",
    "Unknown",
    "is_unknown",
]



from caqtus.types.timelane import DigitalTimeLane
from caqtus.utils import serialization
from caqtus.utils.serialization import JSON
from ._extension import TimeLaneExtension
from ...gui.condetrol.timelanes_editor import TimeLaneModel
from ...gui.condetrol.timelanes_editor.digital_lane_editor import (
    DigitalTimeLaneModel,
    DigitalTimeLaneDelegate,
)


def create_digital_lane(number_steps: int) -> DigitalTimeLane:
    return DigitalTimeLane([False] * number_steps)


def create_lane_model(lane, lane_name: str) -> TimeLaneModel:
    model = DigitalTimeLaneModel(lane_name)
    model.set_lane(lane)
    return model


def create_lane_delegate(lane, lane_name: str) -> DigitalTimeLaneDelegate:
    return DigitalTimeLaneDelegate()


def dump_digital_lane(time_lane: DigitalTimeLane):
    return serialization.converters["json"].unstructure(time_lane, DigitalTimeLane)


def load_digital_lane(content: JSON):
    return serialization.converters["json"].structure(content, DigitalTimeLane)


digital_time_lane_extension = TimeLaneExtension(
    label="Digital",
    lane_type=DigitalTimeLane,
    dumper=dump_digital_lane,
    loader=load_digital_lane,
    lane_factory=create_digital_lane,
    lane_model_factory=create_lane_model,
    lane_delegate_factory=create_lane_delegate,
    type_tag="digital",
)



from typing import Any, TypeAlias

import numpy as np
from typing_extensions import TypeIs

from .._array import FloatArray
from ..recoverable_exceptions import InvalidTypeError
from ..units import Quantity

ScalarAnalogValue: TypeAlias = float | Quantity[float]
ArrayAnalogValue: TypeAlias = FloatArray | Quantity[FloatArray]
AnalogValue: TypeAlias = ScalarAnalogValue | ArrayAnalogValue


class NotAnalogValueError(InvalidTypeError):
    pass


class NotQuantityError(InvalidTypeError):
    pass


def is_scalar_analog_value(value: Any) -> TypeIs[ScalarAnalogValue]:
    """Returns True if the value is a scalar analog value, False otherwise."""

    if isinstance(value, float):
        return True

    if isinstance(value, Quantity):
        return isinstance(value.magnitude, float)

    return False


def is_array_analog_value(value: Any) -> TypeIs[ArrayAnalogValue]:
    """Returns True if the value is an array analog value, False otherwise."""

    if isinstance(value, np.ndarray):
        return issubclass(value.dtype.type, np.floating)

    if isinstance(value, Quantity):
        return isinstance(value.magnitude, np.ndarray) and issubclass(
            value.magnitude.dtype.type, np.floating
        )

    return False


def is_analog_value(value: Any) -> TypeIs[AnalogValue]:
    """Returns True if the value is an analog value, False otherwise."""

    return is_scalar_analog_value(value) or is_array_analog_value(value)


def is_quantity(value: Any) -> TypeIs[Quantity]:
    """Returns True if the value is a quantity, False otherwise."""

    return isinstance(value, Quantity)



import functools
from collections.abc import Callable
from typing import Optional, Protocol, TypeVar

from PySide6.QtWidgets import QStyledItemDelegate

from caqtus.types.timelane import TimeLanes, TimeLane
from caqtus.types.timelane._serializer import TimeLaneSerializer
from caqtus.utils import serialization
from ._protocol import CondetrolLaneExtensionProtocol
from .._time_lane_model import TimeLaneModel

L = TypeVar("L", bound=TimeLane)

type LaneFactory[L: TimeLane] = Callable[[int], L]


class LaneDelegateFactory[L: TimeLane](Protocol):
    """A factory for lane delegates."""

    def __call__(
        self,
        lane: L,
        lane_name: str,
    ) -> Optional[QStyledItemDelegate]:
        """Create a delegate for the lane passed as argument."""
        ...


class LaneModelFactory[L: TimeLane](Protocol):
    def __call__(
        self,
        lane: L,
        lane_name: str,
    ) -> TimeLaneModel[L]:
        """Create a delegate for the lane passed as argument."""
        ...


class CondetrolLaneExtension(CondetrolLaneExtensionProtocol):
    def __init__(self):
        self.get_lane_delegate = functools.singledispatch(default_lane_delegate_factory)
        self.get_lane_model = functools.singledispatch(default_lane_model_factory)
        self._lane_factories: dict[str, LaneFactory] = {}
        self._lane_serializer = TimeLaneSerializer()

    def set_lane_serializer(self, serializer: TimeLaneSerializer) -> None:
        self._lane_serializer = serializer

    def register_lane_factory(self, lane_label: str, factory: LaneFactory) -> None:
        self._lane_factories[lane_label] = factory

    def register_lane_delegate_factory[
        L: TimeLane
    ](self, lane_type: type[L], factory: LaneDelegateFactory[L]) -> None:
        self.get_lane_delegate.register(  # pyright: ignore[reportFunctionMemberAccess]
            lane_type
        )(factory)

    def register_lane_model_factory[
        L: TimeLane
    ](self, lane_type: type[L], factory: LaneModelFactory[L]) -> None:
        self.get_lane_model.register(  # pyright: ignore[reportFunctionMemberAccess]
            lane_type
        )(factory)

    def available_new_lanes(self) -> set[str]:
        return set(self._lane_factories.keys())

    def create_new_lane(self, lane_label: str, steps: int) -> TimeLane:
        lane = self._lane_factories[lane_label](steps)
        if not isinstance(lane, TimeLane):
            raise TypeError(f"Expected a TimeLane, got {type(lane)}.")
        if len(lane) != steps:
            raise ValueError(
                f"Expected a lane with {steps} steps, got {len(lane)} steps."
            )
        return lane

    def unstructure_time_lanes(self, time_lanes: TimeLanes) -> serialization.JsonDict:

        return self._lane_serializer.unstructure_time_lanes(time_lanes)

    def structure_time_lanes(self, content: serialization.JsonDict) -> TimeLanes:
        return self._lane_serializer.structure_time_lanes(content)


def default_lane_model_factory(lane, name: str) -> TimeLaneModel:
    if not isinstance(lane, TimeLane):
        raise TypeError(f"Expected a TimeLane, got {type(lane)}.")

    raise NotImplementedError(f"Don't know how to provide a model for {type(lane)}.")


def default_lane_delegate_factory(
    lane,
    lane_name: str,
) -> None:
    return None



from collections.abc import Iterable

import attrs
import numpy as np

from ._roi import ROI


@attrs.define
class RotatedRectangularROI(ROI):
    center_x: float = attrs.field(
        converter=float,
        on_setattr=attrs.setters.convert,
    )
    center_y: float = attrs.field(
        converter=float,
        on_setattr=attrs.setters.convert,
    )
    width: float = attrs.field(
        converter=float,
        validator=attrs.validators.ge(0),
        on_setattr=attrs.setters.pipe(attrs.setters.convert, attrs.setters.validate),
    )
    height: float = attrs.field(
        converter=float,
        validator=attrs.validators.ge(0),
        on_setattr=attrs.setters.pipe(attrs.setters.convert, attrs.setters.validate),
    )
    angle: float = attrs.field(
        converter=float,
        on_setattr=attrs.setters.convert,
    )

    def get_mask(self) -> np.ndarray:
        x, y = np.mgrid[0 : self.original_width, 0 : self.original_height]

        rotated_x = np.cos(self.angle) * (x - self.center_x) + np.sin(self.angle) * (
            y - self.center_y
        )
        rotated_y = np.cos(self.angle) * (y - self.center_y) - np.sin(self.angle) * (
            x - self.center_x
        )

        mask = np.logical_and(
            np.abs(rotated_x) < self.width / 2, np.abs(rotated_y) < self.height / 2
        )
        return mask

    def get_indices(self) -> tuple[Iterable[int], Iterable[int]]:
        x, y = np.mgrid[0 : self.original_width, 0 : self.original_height]
        mask = self.get_mask()
        return x[mask], y[mask]



from __future__ import annotations

from typing import Optional, Literal, assert_never

import anyio
import attrs
from PySide6.QtCore import Signal, Qt
from PySide6.QtGui import QIcon, QColor, QPalette, QKeySequence, QUndoGroup
from PySide6.QtWidgets import (
    QWidget,
    QToolBar,
    QStackedWidget,
    QLabel,
    QHBoxLayout,
    QApplication,
)
from typing_extensions import assert_type

from caqtus.session import (
    PathNotFoundError,
    SequenceNotEditableError,
    PathIsNotSequenceError,
    State as SequenceState,
)
from caqtus.session import (
    PureSequencePath,
    ExperimentSession,
    AsyncExperimentSession,
    TracebackSummary,
)
from caqtus.session._sequence_collection import (
    SequenceNotCrashedError,
    SequenceNotLaunchedError,
)
from caqtus.session._session_maker import StorageManager
from caqtus.types.iteration import IterationConfiguration
from caqtus.types.iteration.steps_configurations import StepsConfiguration
from caqtus.types.parameter import ParameterNamespace
from caqtus.types.timelane import TimeLanes
from caqtus.utils.result import is_failure_type
from .sequence_widget_ui import Ui_SequenceWidget
from .._icons import get_icon
from .._parameter_tables_editor import ParameterNamespaceEditor
from .._sequence_iteration_editors import StepsIterationEditor
from ..timelanes_editor import TimeLanesEditor
from ..timelanes_editor.extension import CondetrolLaneExtensionProtocol
from ..._common.exception_tree import ExceptionDialog

type State = SequenceNotSet | DraftSequence | NotEditableSequence | CrashedSequence

type LiveState = (
    SequenceNotSet | _DraftSequence | _NotEditableSequence | _CrashedSequence
)


@attrs.frozen
class SequenceNotSet:
    pass


@attrs.frozen
class SequenceSetBase:
    """Base class for sequence states that are set.

    Attributes:
        sequence_path: The path of the sequence.
        iterations: The iteration configuration of the sequence.
        time_lanes: The time lanes of the sequence.
        parameters: The global parameters of the sequence.
    """

    sequence_path: PureSequencePath
    iterations: StepsConfiguration
    time_lanes: TimeLanes
    parameters: ParameterNamespace


@attrs.frozen
class DraftSequence(SequenceSetBase):
    """A sequence that is in the draft state."""

    pass


@attrs.frozen
class NotEditableSequence(SequenceSetBase):
    pass


@attrs.frozen
class CrashedSequence(SequenceSetBase):
    traceback: TracebackSummary


@attrs.frozen
class _LiveStateBase:
    parent: SequenceWidget


@attrs.frozen
class _SetSequenceBase(_LiveStateBase):
    sequence_path: PureSequencePath

    def set_fresh_state(self, state: State) -> None:
        self.parent.set_fresh_state(state)

    def get_parameters(self) -> ParameterNamespace:
        return self.parent.parameters_editor.get_parameters()

    def update_global_parameters(self, parameters: ParameterNamespace) -> None:
        self.parent.iteration_editor.set_available_parameter_names(parameters.names())
        self.parent.parameters_editor.set_parameters(parameters)

    def get_iterations(self) -> IterationConfiguration:
        return self.parent.iteration_editor.get_iteration()

    def set_iterations(self, iterations: StepsConfiguration) -> None:
        self.parent.iteration_editor.set_iteration(iterations)

    def get_time_lanes(self) -> TimeLanes:
        return self.parent.time_lanes_editor.get_time_lanes()

    def set_time_lanes(self, time_lanes: TimeLanes) -> None:
        self.parent.time_lanes_editor.set_time_lanes(time_lanes)


@attrs.frozen
class _DraftSequence(_SetSequenceBase):
    def has_time_lanes_uncommitted_edits(self) -> bool:
        return self.parent.time_lanes_editor.has_uncommitted_edits()

    def time_lanes_edits_committed(self) -> None:
        self.parent.time_lanes_editor.commit_edits()

    def has_iteration_uncommitted_edits(self) -> bool:
        return self.parent.iteration_editor.has_uncommitted_edits()

    def iteration_edits_committed(self) -> None:
        self.parent.iteration_editor.commit_edits()


@attrs.frozen
class _NotEditableSequence(_SetSequenceBase):
    pass


@attrs.frozen
class _CrashedSequence(_SetSequenceBase):
    traceback: TracebackSummary

    def get_traceback(self) -> TracebackSummary:
        return self.traceback

    def set_traceback(self, traceback: TracebackSummary) -> None:
        self.parent._state = _CrashedSequence(
            self.parent, sequence_path=self.sequence_path, traceback=traceback
        )


class SequenceWidget(QWidget, Ui_SequenceWidget):
    """Widget for editing sequence parameters, iterations and time lanes.

    This widget is a tab widget with three tabs: one for defining initial parameters,
    one for editing how the parameters should be iterated over for the sequence, and
    one for editing the time lanes that specify how a given shot should be executed.

    This widget is (optionally) associated with a sequence and displays the iteration
    configuration and time lanes for that sequence.
    If the widget is not associated with a sequence, it will hide itself.

    When associated with a sequence, the widget is constantly watching the state of the
    sequence.
    If the sequence is not in the draft state, the iteration editor and time lanes
    editor
     will become read-only.
    If the sequence is in the draft state, the iteration editor and time lanes editor
    will become editable and any change will be saved.
    """

    sequence_start_requested = Signal(PureSequencePath)
    """A signal emitted when the start sequence button is clicked."""

    def __init__(
        self,
        extension: CondetrolLaneExtensionProtocol,
        parent: Optional[QWidget] = None,
    ):
        """Initializes the sequence widget.

        The sequence widget will initially be associated with no sequence.

        Args:
            extension: The extension that provides time lanes customization.
            parent: The parent widget.
        """

        super().__init__(parent)
        self.setupUi(self)

        self._state: LiveState = SequenceNotSet()
        self.parameters_editor = ParameterNamespaceEditor(self)
        self.parameters_editor.set_read_only(True)
        self.iteration_editor = StepsIterationEditor(self)
        self.time_lanes_editor = TimeLanesEditor(extension, {}, self)
        self.tabWidget.clear()
        self.tabWidget.addTab(self.parameters_editor, "&Globals")
        self.tabWidget.addTab(self.iteration_editor, "&Parameters")
        self.tabWidget.addTab(self.time_lanes_editor, "Time &lanes")

        # undo
        self.undoView.setCleanIcon(
            get_icon("mdi6.content-save", color=Qt.GlobalColor.gray)
        )
        self.undo_group = QUndoGroup(self)
        self.undo_group.addStack(self.parameters_editor.undo_stack)
        self.undo_group.addStack(self.time_lanes_editor.undo_stack)
        self.undo_group.addStack(self.iteration_editor.undo_stack)
        self.tabWidget.currentChanged.connect(self._on_tab_changed)
        self.undo_group.setActiveStack(self.time_lanes_editor.undo_stack)
        self.undoView.setGroup(self.undo_group)
        self.undoView.setVisible(False)

        self.tool_bar = QToolBar(self)
        self.status_widget = IconLabel(icon_position="left")
        self.warning_action = self.tool_bar.addAction(
            get_icon("mdi6.alert", color=QColor(205, 22, 17)), "warning"
        )
        self.warning_action.triggered.connect(self._on_warning_action_triggered)
        self.warning_action.setToolTip("Error")

        self.tool_bar.addWidget(self.status_widget)
        self.start_sequence_action = self.tool_bar.addAction(
            get_icon("start", color=Qt.GlobalColor.darkGreen), "start"
        )
        self.start_sequence_action.triggered.connect(self._on_start_sequence_requested)
        self.start_sequence_action.setShortcut(QKeySequence("F5"))
        self.tool_bar.addSeparator()
        undo_action = self.undo_group.createUndoAction(self)
        undo_icon = get_icon("mdi6.undo")
        undo_action.setIcon(undo_icon)
        undo_action.setShortcut(QKeySequence("Ctrl+Z"))
        self.tool_bar.addAction(undo_action)
        redo_action = self.undo_group.createRedoAction(self)
        redo_icon = get_icon("mdi6.redo")
        redo_action.setIcon(redo_icon)
        redo_action.setShortcut(QKeySequence("Ctrl+Y"))
        self.tool_bar.addAction(redo_action)

        self.verticalLayout.insertWidget(0, self.tool_bar)
        self.stacked = QStackedWidget(self)
        self.stacked.addWidget(QWidget(self))
        self.stacked.addWidget(self.iteration_editor.toolbar)
        self.stacked.addWidget(self.time_lanes_editor.toolbar)
        self.stacked.setCurrentIndex(0)
        self.tool_bar.addSeparator()
        self.tool_bar.addWidget(self.stacked)

        self.tabWidget.currentChanged.connect(self.stacked.setCurrentIndex)

        self._exception_dialog = ExceptionDialog(self)
        self.set_fresh_state(SequenceNotSet())

    def _on_tab_changed(self, index: int) -> None:
        stacks = [
            self.parameters_editor.undo_stack,
            self.iteration_editor.undo_stack,
            self.time_lanes_editor.undo_stack,
        ]
        self.undo_group.setActiveStack(stacks[index])

    def set_fresh_state(self, state: State) -> None:
        match state:
            case SequenceNotSet():
                self.setEnabled(False)
                self.setVisible(False)
                self.start_sequence_action.setEnabled(False)
                new_state = SequenceNotSet()
            case (
                DraftSequence() | NotEditableSequence() | CrashedSequence() as set_state
            ):
                self.iteration_editor.set_iteration(set_state.iterations)
                self.time_lanes_editor.set_time_lanes(set_state.time_lanes)
                self.parameters_editor.set_parameters(set_state.parameters)
                path = set_state.sequence_path
                self.setEnabled(True)
                self.setVisible(True)
                match set_state:
                    case DraftSequence():
                        self.start_sequence_action.setEnabled(True)
                        self.time_lanes_editor.set_read_only(False)
                        self.iteration_editor.set_read_only(False)
                        self.warning_action.setVisible(False)
                        self._set_status_widget(path, True)
                        new_state = _DraftSequence(self, sequence_path=path)
                    case NotEditableSequence():
                        self.start_sequence_action.setEnabled(False)
                        self.time_lanes_editor.set_read_only(True)
                        self.iteration_editor.set_read_only(True)
                        self.warning_action.setVisible(False)
                        self._set_status_widget(path, False)
                        new_state = _NotEditableSequence(self, sequence_path=path)
                    case CrashedSequence(traceback=tb):
                        self.start_sequence_action.setEnabled(False)
                        self.time_lanes_editor.set_read_only(True)
                        self.iteration_editor.set_read_only(True)
                        self.warning_action.setVisible(True)
                        self._set_status_widget(path, False)
                        new_state = _CrashedSequence(
                            self, sequence_path=path, traceback=tb
                        )
                    case _:
                        assert_never(set_state)
            case _:
                assert_never(state)
        self._state = new_state

    def get_current_state(self) -> LiveState:
        return self._state

    def _on_warning_action_triggered(self) -> None:
        """Display the sequence traceback in a dialog."""

        assert isinstance(self._state, _CrashedSequence)

        color = QApplication.palette().color(QPalette.ColorRole.Accent).name()
        self._exception_dialog.set_message(
            f"An error occurred while running the sequence "
            f"<b><font color='{color}'>{self._state.sequence_path}</font></b>."
        )
        traceback = self._state.traceback
        self._exception_dialog.set_exception(traceback)
        self._exception_dialog.show()

    def _set_status_widget(self, path: PureSequencePath, editable: bool) -> None:
        text = " > ".join(path.parts)
        color = self.palette().text().color()
        if editable:
            icon = get_icon("editable-sequence", color=color)
        else:
            icon = get_icon("read-only-sequence", color=color)
        self.status_widget.set_text(text)
        self.status_widget.set_icon(icon)

    def _on_start_sequence_requested(self):
        assert isinstance(self._state, _DraftSequence)
        self.sequence_start_requested.emit(self._state.sequence_path)

    async def exec_async(self, storage_manager: StorageManager):
        """Run a synchronization loop to keep to editor in sync with the storage."""

        await update(self, storage_manager)


async def update(widget: SequenceWidget, storage_manager: StorageManager) -> None:
    while True:
        async with storage_manager.async_session() as session:
            await synchronize_editor_and_storage(widget, session)
        await anyio.sleep(100e-3)


async def synchronize_editor_and_storage(
    editor: SequenceWidget, session: AsyncExperimentSession
) -> bool:
    """Synchronize what is displayed in the editor with the session storage.

    Returns:
        True if there was some difference between the editor and the storage, and
        the changes were resolved.
        False otherwise.
    """

    editor_state = editor.get_current_state()
    match editor_state:
        case SequenceNotSet():
            return False
        case (
            _DraftSequence(sequence_path=path)
            | _NotEditableSequence(sequence_path=path)
            | _CrashedSequence(sequence_path=path)
        ):
            storage_state = await _query_state_async(path, session)
            if editor_state != editor.get_current_state():
                # Could be that the editor state changed while fetching the data from
                # the storage.
                # In this case we relaunch the synchronization.
                return await synchronize_editor_and_storage(editor, session)
            else:
                if isinstance(storage_state, SequenceNotSet):
                    editor.set_fresh_state(storage_state)
                    return True
                return await synchronize_editor_and_storage_sequence_data(
                    editor_state, session, storage_state
                )

        case _:
            assert_never(editor_state)


async def synchronize_editor_and_storage_sequence_data(
    editor_state: _DraftSequence | _NotEditableSequence | _CrashedSequence,
    session: AsyncExperimentSession,
    storage_state: DraftSequence | NotEditableSequence | CrashedSequence,
) -> bool:
    match editor_state:
        case _DraftSequence() as editor_state:
            return await synchronize_draft_sequence(
                editor_state, session, storage_state
            )
        case _NotEditableSequence() as editor_state:
            return await synchronize_not_editable_sequence(editor_state, storage_state)
        case _CrashedSequence() as editor_state:
            return await synchronize_crashed_sequence(editor_state, storage_state)
        case _:
            assert_never(editor_state)


async def synchronize_draft_sequence(
    editor_state: _DraftSequence, session: AsyncExperimentSession, state: State
) -> bool:
    match state:
        case SequenceNotSet() | NotEditableSequence() | CrashedSequence():
            editor_state.set_fresh_state(state)
            return True
        case DraftSequence():
            changes = False

            if editor_state.get_parameters() != state.parameters:
                editor_state.update_global_parameters(state.parameters)
                changes = True

            if editor_state.has_iteration_uncommitted_edits():
                iterations = editor_state.get_iterations()
                editor_state.iteration_edits_committed()
                result = await session.sequences.set_iteration_configuration(
                    editor_state.sequence_path, iterations
                )
                assert not is_failure_type(
                    result,
                    (
                        PathNotFoundError,
                        PathIsNotSequenceError,
                        SequenceNotEditableError,
                    ),
                ), "Path should exists in the session and be an editable sequence."
                assert_type(result, None)
                changes = True
            elif editor_state.get_iterations() != state.iterations:
                editor_state.set_iterations(state.iterations)
                changes = True

            editor_time_lanes = editor_state.get_time_lanes()
            if editor_state.has_time_lanes_uncommitted_edits():
                editor_state.time_lanes_edits_committed()
                result = await session.sequences.set_time_lanes(
                    editor_state.sequence_path, editor_time_lanes
                )
                assert not is_failure_type(
                    result,
                    (
                        PathNotFoundError,
                        PathIsNotSequenceError,
                        SequenceNotEditableError,
                    ),
                ), "Path should exists in the session and be an editable sequence."
                assert_type(result, None)
                changes = True
            elif editor_time_lanes != state.time_lanes:
                editor_state.set_time_lanes(state.time_lanes)
                changes = True
            return changes
        case _:
            assert_never(state)


async def synchronize_not_editable_sequence(
    editor_state: _NotEditableSequence, state: State
) -> bool:
    match state:
        case SequenceNotSet() | DraftSequence() | CrashedSequence():
            editor_state.set_fresh_state(state)
            return True
        case NotEditableSequence():
            changes = False
            if editor_state.get_parameters() != state.parameters:
                editor_state.update_global_parameters(state.parameters)
                changes = True

            if editor_state.get_iterations() != state.iterations:
                editor_state.set_iterations(state.iterations)
                changes = True

            if editor_state.get_time_lanes() != state.time_lanes:
                editor_state.set_time_lanes(state.time_lanes)
                changes = True
            return changes
        case _:
            assert_never(state)


async def synchronize_crashed_sequence(
    editor_state: _CrashedSequence, state: State
) -> bool:
    match state:
        case SequenceNotSet() | DraftSequence() | NotEditableSequence():
            editor_state.set_fresh_state(state)
            return True
        case CrashedSequence():
            changes = False
            if editor_state.get_parameters() != state.parameters:
                editor_state.update_global_parameters(state.parameters)
                changes = True

            if editor_state.get_iterations() != state.iterations:
                editor_state.set_iterations(state.iterations)
                changes = True

            if editor_state.get_time_lanes() != state.time_lanes:
                editor_state.set_time_lanes(state.time_lanes)
                changes = True

            if editor_state.get_traceback() != state.traceback:
                editor_state.set_traceback(state.traceback)
                changes = True
            return changes
        case _:
            assert_never(state)


def _query_state_sync(path: PureSequencePath, session: ExperimentSession) -> State:
    is_sequence_result = session.sequences.is_sequence(path)
    if is_failure_type(is_sequence_result, PathNotFoundError):
        return SequenceNotSet()
    else:
        if is_sequence_result.result():
            return _query_sequence_state_sync(path, session)
        else:
            return SequenceNotSet()


def _query_sequence_state_sync(
    path: PureSequencePath, session: ExperimentSession
) -> DraftSequence | NotEditableSequence | CrashedSequence:
    state_result = session.sequences.get_state(path)
    assert not is_failure_type(
        state_result, (PathNotFoundError, PathIsNotSequenceError)
    ), "Path should exists in the session and be a sequence."
    state = state_result.content()
    iterations = session.sequences.get_iteration_configuration(path)
    if not isinstance(iterations, StepsConfiguration):
        raise NotImplementedError("Only steps iterations are supported.")
    time_lanes = session.sequences.get_time_lanes(path)
    assert not is_failure_type(
        time_lanes, (PathNotFoundError, PathIsNotSequenceError)
    ), "Path should exists in the session and be a sequence."

    if state.is_editable():
        parameters = session.get_global_parameters()
        return DraftSequence(
            path, iterations=iterations, time_lanes=time_lanes, parameters=parameters
        )
    else:
        parameters_result = session.sequences.get_global_parameters(path)
        assert not is_failure_type(
            parameters_result,
            (PathNotFoundError, PathIsNotSequenceError, SequenceNotLaunchedError),
        ), "Path should exists in the session and be a launched sequence."
        parameters = parameters_result.content()
        if state == SequenceState.CRASHED:
            traceback_summary_result = session.sequences.get_exception(path)
            assert not is_failure_type(
                traceback_summary_result,
                (PathNotFoundError, PathIsNotSequenceError, SequenceNotCrashedError),
            ), "Path should exists in the session and be a crashed sequence."
            traceback_summary = traceback_summary_result.content()
            if traceback_summary is None:
                error = RuntimeError("Sequence crashed but no traceback available.")
                traceback_summary = TracebackSummary.from_exception(error)
            return CrashedSequence(
                path,
                iterations=iterations,
                time_lanes=time_lanes,
                parameters=parameters,
                traceback=traceback_summary,
            )
        else:
            return NotEditableSequence(
                path,
                iterations=iterations,
                time_lanes=time_lanes,
                parameters=parameters,
            )


async def _query_state_async(
    path: PureSequencePath, session: AsyncExperimentSession
) -> State:
    is_sequence_result = await session.sequences.is_sequence(path)
    if is_failure_type(is_sequence_result, PathNotFoundError):
        return SequenceNotSet()
    else:
        if is_sequence_result.result():
            return await _query_sequence_state_async(path, session)
        else:
            return SequenceNotSet()


async def _query_sequence_state_async(
    path: PureSequencePath, session: AsyncExperimentSession
) -> DraftSequence | NotEditableSequence | CrashedSequence:
    state_result = await session.sequences.get_state(path)
    assert not is_failure_type(
        state_result, (PathNotFoundError, PathIsNotSequenceError)
    ), "Path should exists in the session and be a sequence."
    state = state_result.content()
    iterations = await session.sequences.get_iteration_configuration(path)
    if not isinstance(iterations, StepsConfiguration):
        raise NotImplementedError("Only steps iterations are supported.")
    time_lanes = await session.sequences.get_time_lanes(path)
    assert not is_failure_type(
        time_lanes, (PathNotFoundError, PathIsNotSequenceError)
    ), "Path should exists in the session and be a sequence."

    if state.is_editable():
        parameters = await session.get_global_parameters()
        return DraftSequence(
            path, iterations=iterations, time_lanes=time_lanes, parameters=parameters
        )
    else:
        parameters_result = await session.sequences.get_global_parameters(path)
        assert not is_failure_type(
            parameters_result,
            (PathNotFoundError, PathIsNotSequenceError, SequenceNotLaunchedError),
        ), "Path should exists in the session and be a launched sequence."
        parameters = parameters_result.content()
        if state == SequenceState.CRASHED:
            traceback_summary_result = await session.sequences.get_traceback_summary(
                path
            )
            assert not is_failure_type(
                traceback_summary_result,
                (PathNotFoundError, PathIsNotSequenceError, SequenceNotCrashedError),
            ), "Path should exists in the session and be a crashed sequence."
            traceback_summary = traceback_summary_result.content()
            if traceback_summary is None:
                error = RuntimeError("Sequence crashed but no traceback available.")
                traceback_summary = TracebackSummary.from_exception(error)
            return CrashedSequence(
                path,
                iterations=iterations,
                time_lanes=time_lanes,
                parameters=parameters,
                traceback=traceback_summary,
            )
        else:
            return NotEditableSequence(
                path,
                iterations=iterations,
                time_lanes=time_lanes,
                parameters=parameters,
            )


class IconLabel(QWidget):
    def __init__(
        self,
        parent: Optional[QWidget] = None,
        icon_position: Literal["left", "right"] = "left",
    ):
        super().__init__(parent)
        self._label = QLabel()
        self._icon = QLabel()
        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        if icon_position == "left":
            layout.addWidget(self._icon)
            layout.addWidget(self._label)
        else:
            layout.addWidget(self._label)
            layout.addWidget(self._icon)
        self.setLayout(layout)

    def set_text(self, text: str):
        self._label.setText(text)

    def set_icon(self, icon: Optional[QIcon]):
        if icon is None:
            self._icon.clear()
        else:
            self._icon.setPixmap(icon.pixmap(20, 20))



"""Contains utility functions for copying data between sessions.

All functions defined in this package are independent on the implementation of the
session.
"""

from ._copy import copy_path

__all__ = ["copy_path"]



from __future__ import annotations

import abc
from typing import Optional, Any

from PySide6.QtCore import (
    QObject,
    Qt,
    QSettings,
)
from PySide6.QtGui import QAction, QBrush, QColor
from PySide6.QtWidgets import QMenu, QColorDialog

import caqtus.gui.qtutil.qabc as qabc
from caqtus.types.timelane import TimeLane
from ._time_lane_model import TimeLaneModel


class ColoredTimeLaneModel[L: TimeLane](TimeLaneModel[L], metaclass=qabc.QABCMeta):
    """A time lane model that can be colored.

    Instances of this class can be used to color the cells in a lane.
    They have the attribute :attr:`_brush` that is optionally a :class:`QBrush` that
    can be used to color the cells in the lane.
    """

    # ruff: noqa: N802

    def __init__(
        self,
        name: str,
        lane: L,
        parent: Optional[QObject] = None,
    ):
        super().__init__(name, lane, parent)
        self._brush: Optional[QBrush] = None

        color = QSettings().value(f"lane color/{self.name()}", None)
        if color is not None:
            self._brush = QBrush(color)
        else:
            self._brush = None

    @abc.abstractmethod
    def data(self, index, role: int = Qt.ItemDataRole.DisplayRole) -> Any:
        """Returns its brush for the `Qt.ItemDataRole.ForegroundRole` role."""

        if not index.isValid():
            return None
        if role == Qt.ItemDataRole.ForegroundRole:
            return self._brush

    def headerData(
        self,
        section: int,
        orientation: Qt.Orientation,
        role: int = Qt.ItemDataRole.DisplayRole,
    ):
        if role == Qt.ItemDataRole.ForegroundRole:
            return self._brush
        return super().headerData(section, orientation, role)

    def get_header_context_actions(self) -> list[QAction | QMenu]:
        action = QAction("Change color")
        action.triggered.connect(lambda: self._change_color())
        return [action]

    def _change_color(self):
        if self._brush is None:
            color = QColorDialog.getColor(title=f"Select color for {self.name()}")
        else:
            color = QColorDialog.getColor(
                self._brush.color(), title=f"Select color for {self.name()}"
            )
        if color.isValid():
            self.setHeaderData(
                0, Qt.Orientation.Horizontal, color, Qt.ItemDataRole.ForegroundRole
            )

    def setHeaderData(
        self, section, orientation, value, role: int = Qt.ItemDataRole.EditRole
    ):
        change = False
        if (
            orientation == Qt.Orientation.Horizontal
            and role == Qt.ItemDataRole.ForegroundRole
        ):
            if isinstance(value, QColor):
                self._brush = QBrush(value)
                settings = QSettings()
                settings.setValue(f"lane color/{self.name()}", value)
                change = True
            elif value is None:
                self._brush = None
                settings = QSettings()
                settings.remove(f"lane color/{self.name()}")
                change = True
        if change:
            self.headerDataChanged.emit(orientation, section, section)
            return True

        return super().setHeaderData(section, orientation, value, role)



from collections.abc import Callable, Awaitable
from typing import Optional

import anyio
from PySide6 import QtWidgets

from caqtus.gui.qtutil import temporary_widget


async def blocking_call[
    **P, T
](
    parent: Optional[QtWidgets.QWidget],
    msg: str,
    fun: Callable[P, Awaitable[T]],
    *args: P.args,
    **kwargs: P.kwargs,
) -> None:
    """Run an async function while blocking the application with a message box.

    When the function terminates, the message box is closed and the user can resume
    interacting with the application.

    If the cancel button of the message box is clicked, the function will be cancelled.
    The message box will be closed immediately, but the function will continue to run
    until it reaches a cancellation point.

    It is the responsibility of the function to check for cancellation and handle it
    appropriately.

    Args:
        parent: The parent widget for the message box.
        msg: The message to display in the message box.
        fun: The async function to run.
        args: The arguments to pass to the function.
        kwargs: The keyword arguments to pass to the function
    """

    cancel_scope = anyio.CancelScope()

    def on_finished(result: QtWidgets.QDialog.DialogCode):
        if result == QtWidgets.QMessageBox.StandardButton.Cancel:
            cancel_scope.cancel()

    with temporary_widget(QtWidgets.QMessageBox(parent)) as message_box:

        async def wrapped():
            with cancel_scope:
                ret = await fun(*args, **kwargs)
                message_box.accept()
                return ret

        message_box.setText(msg)
        message_box.setIcon(QtWidgets.QMessageBox.Icon.Information)
        message_box.setStandardButtons(QtWidgets.QMessageBox.StandardButton.Cancel)
        message_box.finished.connect(on_finished)

        async with anyio.create_task_group() as tg:
            tg.start_soon(wrapped)
            message_box.open()



"""This module define the type of data that can be saved for a shot."""

from ._data_label import DataLabel, is_data_label
from ._data_type import Data, Array, StructuredData
from ._is_data import is_data

__all__ = ["Data", "Array", "StructuredData", "is_data", "DataLabel", "is_data_label"]



import abc
from typing import Protocol

from caqtus.utils.serialization import JsonDict
from ..timelane import TimeLane, TimeLanes


class TimeLaneSerializerProtocol(Protocol):
    """Defines how to (un)structure time lanes."""

    @abc.abstractmethod
    def dump(self, lane: TimeLane) -> JsonDict:
        """Serialize a single time lane to a JSON dictionary."""

        ...

    @abc.abstractmethod
    def load(self, data: JsonDict) -> TimeLane:
        """Construct a single time lane from a JSON dictionary."""

        ...

    @abc.abstractmethod
    def unstructure_time_lanes(self, time_lanes: TimeLanes) -> JsonDict: ...

    @abc.abstractmethod
    def structure_time_lanes(self, content: JsonDict) -> TimeLanes: ...



from collections.abc import Mapping
from copy import deepcopy
from typing import Generic, TypeVar, Self, Optional

from caqtus.types.variable_name import DottedVariableName
from .._parameter_namespace import VariableNamespace

T = TypeVar("T")


class StepContext(Generic[T]):
    """Immutable context that contains the variables of a given step."""

    def __init__(
        self, initial_variables: Optional[Mapping[DottedVariableName, T]] = None
    ) -> None:
        self._variables = VariableNamespace[T]()
        if initial_variables is not None:
            self._variables.update(dict(initial_variables))

    def clone(self) -> Self:
        return deepcopy(self)

    def update_variable(self, name: DottedVariableName, value: T) -> Self:
        clone = self.clone()
        clone._variables.update({name: value})
        return clone

    @property
    def variables(self) -> VariableNamespace[T]:
        return deepcopy(self._variables)



import contextlib
import ctypes
import logging.config
import platform
import warnings
from collections.abc import Callable
from typing import Optional, assert_never, Concatenate

from typing_extensions import deprecated

from caqtus.experiment_control.manager import (
    LocalExperimentManager,
    RemoteExperimentManagerClient,
    RemoteExperimentManagerServer,
)

# noinspection PyProtectedMember
from caqtus.gui.condetrol._condetrol import Condetrol
from caqtus.session.sql import (
    PostgreSQLConfig,
    PostgreSQLStorageManager,
    SQLiteConfig,
    SQLiteStorageManager,
    SQLStorageManager,
)
from ._caqtus_extension import CaqtusExtension
from .device_extension import DeviceExtension
from .time_lane_extension import TimeLaneExtension
from ..device.configuration import DeviceServerName
from ..device.remote import Server, RPCConfiguration
from ..experiment_control import ExperimentManager
from ..experiment_control.manager import (
    ExperimentManagerConnection,
    LocalExperimentManagerConfiguration,
    RemoteExperimentManagerConfiguration,
)
from ..experiment_control.sequence_execution import ShotRetryConfig
from ..session import ExperimentSession, StorageManager
from ..session.sql._serializer import SerializerProtocol


class Experiment:
    """Configure parameters and register extensions for a specific experiment.

    There should be only a single instance of this class in the entire application.
    It is used to configure the experiment and knows how to launch the different
    components of the application after it has been configured.

    Args:
        storage_config: The configuration of the storage backend to be used to store the
            data of the experiment.
    """

    def __init__(
        self, storage_config: PostgreSQLConfig | SQLiteConfig | None = None
    ) -> None:
        if storage_config is None:
            warnings.warn(
                "A storage configuration should be passed when initializing the "
                "experiment.",
                DeprecationWarning,
                stacklevel=2,
            )
        self._storage_config: PostgreSQLConfig | SQLiteConfig | None = storage_config
        self._extension = CaqtusExtension()
        self._experiment_manager: Optional[LocalExperimentManager] = None
        self._experiment_manager_location: ExperimentManagerConnection = (
            LocalExperimentManagerConfiguration()
        )
        self._shot_retry_config: Optional[ShotRetryConfig] = None

    def setup_default_extensions(self) -> None:
        """Register some commonly used extensions to this experiment.

        This method registers the following extensions:

        * digital time lanes
        * analog time lanes
        * camera time lanes
        """

        from caqtus.extension.time_lane_extension import (
            digital_time_lane_extension,
            analog_time_lane_extension,
            camera_time_lane_extension,
        )

        self.register_time_lane_extension(digital_time_lane_extension)
        self.register_time_lane_extension(analog_time_lane_extension)
        self.register_time_lane_extension(camera_time_lane_extension)

    @deprecated(
        "Pass the configuration directly to the Experiment() constructor instead."
    )
    def configure_storage(
        self, storage_config: PostgreSQLConfig | SQLiteConfig
    ) -> None:
        """Configure the storage backend to be used by the application.

        After this method is called, the application will read and write data and
        configurations to the storage specified.

        It is necessary to call this method before launching the application.

        Warning:
            Calling this method multiple times will overwrite the previous
            configuration.
        """

        if self._storage_config is not None:
            warnings.warn("Storage configuration is being overwritten.", stacklevel=2)
        self._storage_config = storage_config

    def configure_shot_retry(
        self, shot_retry_config: Optional[ShotRetryConfig]
    ) -> None:
        """Configure the shot retry policy to be used when running sequences.

        After this method is called, shots that raise errors will be retried according
        to the policy specified.

        It is necessary to call this method before launching the experiment manager.

        Warning:
            Calling this method multiple times will overwrite the previous
            configuration.
        """

        self._shot_retry_config = shot_retry_config

    def configure_experiment_manager(
        self, location: ExperimentManagerConnection
    ) -> None:
        """Configure the location of the experiment manager with respect to Condetrol.

        The :class:`ExperimentManager` is responsible for running sequences on the
        experiment.

        It can be either running in the same process as the Condetrol application or in
        a separate process.

        This is configured by passing an instance of either
        :class:`LocalExperimentManagerConfiguration` or
        :class:`RemoteExperimentManagerConfiguration`.

        If this method is not called, the experiment manager will be assumed to be
        running in the same local process as the Condetrol application.

        If the experiment manager is configured to run in the same process, it will be
        created when the Condetrol application is launched.
        An issue with this approach is that if the Condetrol application crashes, the
        experiment manager will also stop abruptly, potentially leaving the experiment
        in an undesired state.

        If the experiment manager is configured to run in a separate process, it will be
        necessary to have an experiment manager server running before launching the
        Condetrol application.
        The Condetrol application will then connect to the server and transmit the
        commands to the other process.
        If the Condetrol application crashes, the experiment manager will be unaffected.

        Warning:
            Calling this method multiple times will overwrite the previous
            configuration.
        """

        self._experiment_manager_location = location

    def register_device_extension(self, device_extension: DeviceExtension) -> None:
        """Register a new device extension.

        After this method is called, the device extension will be available to the
        application, both in the device editor tab in Condetrol and while running the
        experiment.
        """

        self._extension.register_device_extension(device_extension)

    def register_time_lane_extension(
        self, time_lane_extension: TimeLaneExtension
    ) -> None:
        """Register a new time lane extension.

        After this method is called, the time lane extension will be available to the
        application, both in the time lane editor tab in Condetrol and while running the
        experiment.
        """

        self._extension.register_time_lane_extension(time_lane_extension)

    def register_device_server(
        self, name: DeviceServerName, config: RPCConfiguration
    ) -> None:
        """Register a new device server.

        After this method is called, the device server will be available to the
        application to connect to devices.
        """

        self._extension.register_device_server_config(name, config)

    def get_storage_manager(self) -> StorageManager:
        """Get the storage manager to be used by the application.

        The storage manager is responsible for interacting with the storage of the
        experiment.
        """

        return self._get_storage_manager(check_schema=True)

    @deprecated("Use get_storage_manager instead.")
    def get_session_maker(self) -> StorageManager:
        return self.get_storage_manager()

    def build_storage_manager[
        T: StorageManager, **P
    ](
        self,
        backend_type: Callable[Concatenate[SerializerProtocol, P], T],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> T:
        """Create and set up a storage manager with the current registered extensions.

        Args:
            backend_type: Defines how the data will be stored.
            *args: The arguments to pass to the storage backend constructor.
            **kwargs: The keyword arguments to pass to the storage backend constructor.
        """

        storage_backend_manager = self._build_storage_manager(
            backend_type,
            *args,
            **kwargs,
        )
        if isinstance(storage_backend_manager, SQLStorageManager):
            storage_backend_manager.check()

        return storage_backend_manager

    def _build_storage_manager[
        T: StorageManager, **P
    ](
        self,
        backend_type: Callable[Concatenate[SerializerProtocol, P], T],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> T:
        return self._extension.create_session_maker(
            backend_type,
            *args,
            **kwargs,
        )

    def _get_storage_manager(self, check_schema: bool = True) -> StorageManager:
        if self._storage_config is None:
            error = RuntimeError("Storage configuration has not been set.")
            error.add_note(
                "Pass the storage configuration to the Experiment() constructor."
            )
            raise error
        if isinstance(self._storage_config, SQLiteConfig):
            storage_manager = self._build_storage_manager(
                SQLiteStorageManager,
                config=self._storage_config,
            )
        elif isinstance(self._storage_config, PostgreSQLConfig):
            storage_manager = self._build_storage_manager(
                PostgreSQLStorageManager,
                config=self._storage_config,
            )
        else:
            assert_never(self._storage_config)
        if check_schema:
            storage_manager.check()
        return storage_manager

    def connect_to_experiment_manager(self) -> ExperimentManager:
        """Connect to the experiment manager."""

        location = self._experiment_manager_location
        if isinstance(location, LocalExperimentManagerConfiguration):
            return self.get_local_experiment_manager()
        elif isinstance(location, RemoteExperimentManagerConfiguration):
            client = RemoteExperimentManagerClient(
                address=(location.address, location.port),
                authkey=bytes(location.authkey, "utf-8"),
            )
            return client.get_experiment_manager()
        else:
            assert_never(location)

    def get_local_experiment_manager(self) -> LocalExperimentManager:
        """Return the local experiment manager.

        This method is used to create an instance of the experiment manager that runs
        in the local process.

        The first time this method is called, the experiment manager will be created.
        If it is called again, the instance previously created will be returned.
        """

        if self._experiment_manager is None:
            self._experiment_manager = LocalExperimentManager(
                session_maker=self.get_storage_manager(),
                device_manager_extension=self._extension.device_manager_extension,
                shot_retry_config=self._shot_retry_config,
            )
        return self._experiment_manager

    def launch_condetrol(self) -> None:
        """Launch the Condetrol application.

        The Condetrol application is the main user interface to the experiment.
        It allows to edit and launch sequences, as well as edit the device
        configurations.
        """

        if platform.system() == "Windows":
            # This is necessary to use the UI icon in the taskbar and not the default
            # Python icon.
            app_id = "caqtus.condetrol"
            ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(app_id)  # type: ignore[reportAttributeAccessIssue]

        app = Condetrol(
            self.get_storage_manager(),
            extension=self._extension.condetrol_extension,
            connect_to_experiment_manager=self.connect_to_experiment_manager,
        )
        try:
            app.run()
        except:
            logging.exception("An error occurred.", exc_info=True)
            raise

    def launch_experiment_server(self) -> None:
        """Launch the experiment server.

        The experiment server is used to run procedures on the experiment manager from a
        remote process.
        """

        if platform.system() == "Windows":
            # This is necessary to use the UI icon in the taskbar and not the default
            # Python icon.
            app_id = "caqtus.experiment_server"
            ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(app_id)  # type: ignore[reportAttributeAccessIssue]

        if not isinstance(
            self._experiment_manager_location, RemoteExperimentManagerConfiguration
        ):
            error = RuntimeError(
                "The experiment manager is not configured to run remotely."
            )
            error.add_note(
                "Please call `configure_experiment_manager` with a remote "
                "configuration."
            )
            raise error

        server = RemoteExperimentManagerServer(
            session_maker=self.get_storage_manager(),
            address=("localhost", self._experiment_manager_location.port),
            authkey=bytes(self._experiment_manager_location.authkey, "utf-8"),
            shot_retry_config=self._shot_retry_config,
            device_manager_extension=self._extension.device_manager_extension,
        )

        with server:
            print("Ready")
            server.serve_forever()

    @staticmethod
    def launch_device_server(
        config: RPCConfiguration, name: str = "device_server"
    ) -> None:
        """Launch a device server in the current process.

        This method will block until the server is stopped.

        Args:
            config: The configuration of the server.
            name: The name of the server. It is used to create the log file.
        """

        if platform.system() == "Windows":
            # This is necessary to use the UI icon in the taskbar and not the default
            # Python icon.
            app_id = "caqtus.device_server"
            ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(app_id)  # type: ignore[reportAttributeAccessIssue]

        with Server(config) as server:
            print("Ready")
            server.wait_for_termination()

    def storage_session(self) -> contextlib.AbstractContextManager[ExperimentSession]:
        """Return a context manager that provides a session to the storage backend.

        A session can be used to access the data stored in the experiment.
        """

        return self.get_storage_manager().session()


def upgrade_database(experiment: Experiment) -> None:
    """Upgrade the database schema of the experiment to the latest version.

    .. Warning::

        It is strongly recommended to back up the database before running this
        function in case something goes wrong.

    Args:
        experiment: The experiment to upgrade the database for.
            It must have been configured with a PostgreSQL storage backend.
    """

    storage_manager = experiment._get_storage_manager(check_schema=False)
    if not isinstance(storage_manager, SQLStorageManager):
        error = RuntimeError("The storage manager is not a SQL storage manager.")
        error.add_note(
            "The upgrade_database method is only available for SQL storage managers."
        )
        raise error
    storage_manager.upgrade()


def stamp_database(experiment: Experiment) -> None:
    """Mark old databases schema with the original revision.

    This should only be called on databases that were created before version 6.3.0.
    """

    from alembic.command import stamp

    storage_manager = experiment._get_storage_manager(check_schema=False)
    if not isinstance(storage_manager, PostgreSQLStorageManager):
        raise RuntimeError("The storage manager is not a PostgreSQL storage manager.")
    config = storage_manager._get_alembic_config()

    stamp(config, "038164d73465")



from __future__ import annotations

from abc import abstractmethod
from typing import Optional, TypeVar, Generic, Callable, NewType

import attrs
from PySide6.QtWidgets import QWidget

import caqtus.gui.qtutil.qabc as qabc
from caqtus.session._shot_id import ShotId
from caqtus.utils.serialization import JSON


class ShotView(QWidget, metaclass=qabc.QABCMeta):
    @abstractmethod
    async def display_shot(self, shot: ShotId) -> None:
        raise NotImplementedError


S = TypeVar("S", bound=JSON)
V = TypeVar("V", bound=ShotView)

ManagerName = NewType("ManagerName", str)


@attrs.define
class ViewManager(Generic[V, S]):
    constructor: Callable[[S], V]
    dumper: Callable[[V], S]
    state_generator: Callable[[QWidget], Optional[tuple[str, S]]]
